@@ -32,7 +32,7 @@ impl<'a> Vm<'a> {
         scopes: Scopes<'a>,
         target: Span,
     ) -> Self {
-        let inspected = target.id().and_then(|id| engine.tracer.inspected(id));
+        let inspected = target.id().and_then(|id| engine.traced.get(id));
         Self { engine, context, flow: None, scopes, inspected }
     }
 
@@ -47,14 +47,14 @@ impl<'a> Vm<'a> {
         if self.inspected == Some(var.span()) {
             self.trace(value.clone());
         }
-        self.scopes.top.define(var.get().clone(), value);
+        self.scopes.top.define_ident(var, value);
     }
 
     /// Trace a value.
     #[cold]
     pub fn trace(&mut self, value: Value) {
         self.engine
-            .tracer
+            .sink
             .value(value.clone(), self.context.styles().ok().map(|s| s.to_map()));
     }
 }
