@@ -18,13 +18,12 @@ use crate::foundations::{
     NativeElement, Recipe, RecipeIndex, Repr, Selector, Str, Style, StyleChain, Styles,
     Value,
 };
-use crate::introspection::{Location, Meta, MetaElem};
+use crate::introspection::Location;
 use crate::layout::{AlignElem, Alignment, Axes, Length, MoveElem, PadElem, Rel, Sides};
-use crate::model::{Destination, EmphElem, StrongElem};
-use crate::realize::{Behave, Behaviour};
+use crate::model::{Destination, EmphElem, LinkElem, StrongElem};
 use crate::syntax::Span;
 use crate::text::UnderlineElem;
-use crate::util::{fat, BitSet, LazyHash};
+use crate::utils::{fat, singleton, LazyHash, SmallBitSet};
 
 /// A piece of document content.
 ///
@@ -73,7 +72,7 @@ use crate::util::{fat, BitSet, LazyHash};
 #[derive(Clone, Hash)]
 #[allow(clippy::derived_hash_with_manual_eq)]
 pub struct Content {
-    /// The partially element-dependant inner data.
+    /// The partially element-dependent inner data.
     inner: Arc<Inner<dyn Bounds>>,
     /// The element's source code location.
     span: Span,
@@ -90,7 +89,7 @@ struct Inner<T: ?Sized + 'static> {
     /// - If bit 0 is set, the element is prepared.
     /// - If bit n is set, the element is guarded against the n-th show rule
     ///   recipe from the top of the style chain (counting from 1).
-    lifecycle: BitSet,
+    lifecycle: SmallBitSet,
     /// The element's raw data.
     elem: LazyHash<T>,
 }
@@ -102,16 +101,16 @@ impl Content {
             inner: Arc::new(Inner {
                 label: None,
                 location: None,
-                lifecycle: BitSet::new(),
+                lifecycle: SmallBitSet::new(),
                 elem: elem.into(),
             }),
             span: Span::detached(),
         }
     }
 
-    /// Creates a new empty sequence content.
+    /// Creates a empty sequence content.
     pub fn empty() -> Self {
-        Self::new(SequenceElem::default())
+        singleton!(Content, SequenceElem::default().pack()).clone()
     }
 
     /// Get the element of this content.
@@ -137,20 +136,26 @@ impl Content {
         self.inner.label
     }
 
-    /// Set the label of the content.
+    /// Attach a label to the content.
     pub fn labelled(mut self, label: Label) -> Self {
-        self.make_mut().label = Some(label);
+        self.set_label(label);
         self
     }
 
-    /// Check whether a show rule recipe is disabled.
-    pub fn is_guarded(&self, index: RecipeIndex) -> bool {
-        self.inner.lifecycle.contains(index.0)
+    /// Set the label of the content.
+    pub fn set_label(&mut self, label: Label) {
+        self.make_mut().label = Some(label);
     }
 
-    /// Whether this content has already been prepared.
-    pub fn is_prepared(&self) -> bool {
-        self.inner.lifecycle.contains(0)
+    /// Assigns a location to the content.
+    ///
+    /// This identifies the content and e.g. makes it linkable by
+    /// `.linked(Destination::Location(loc))`.
+    ///
+    /// Useful in combination with [`Location::variant`].
+    pub fn located(mut self, loc: Location) -> Self {
+        self.set_location(loc);
+        self
     }
 
     /// Set the location of the content.
@@ -158,32 +163,40 @@ impl Content {
         self.make_mut().location = Some(location);
     }
 
+    /// Check whether a show rule recipe is disabled.
+    pub fn is_guarded(&self, index: RecipeIndex) -> bool {
+        self.inner.lifecycle.contains(index.0)
+    }
+
     /// Disable a show rule recipe.
     pub fn guarded(mut self, index: RecipeIndex) -> Self {
         self.make_mut().lifecycle.insert(index.0);
         self
     }
 
+    /// Whether this content has already been prepared.
+    pub fn is_prepared(&self) -> bool {
+        self.inner.lifecycle.contains(0)
+    }
+
     /// Mark this content as prepared.
     pub fn mark_prepared(&mut self) {
         self.make_mut().lifecycle.insert(0);
     }
 
-    /// How this element interacts with other elements in a stream.
-    pub fn behaviour(&self) -> Behaviour {
-        self.with::<dyn Behave>()
-            .map_or(Behaviour::Supportive, Behave::behaviour)
-    }
-
     /// Get a field by ID.
     ///
     /// This is the preferred way to access fields. However, you can only use it
     /// if you have set the field IDs yourself or are using the field IDs
     /// generated by the `#[elem]` macro.
-    pub fn get(&self, id: u8, styles: Option<StyleChain>) -> Option<Value> {
+    pub fn get(
+        &self,
+        id: u8,
+        styles: Option<StyleChain>,
+    ) -> Result<Value, FieldAccessError> {
         if id == 255 {
             if let Some(label) = self.label() {
-                return Some(label.into_value());
+                return Ok(label.into_value());
             }
         }
         match styles {
@@ -196,13 +209,13 @@ impl Content {
     ///
     /// If you have access to the field IDs of the element, use [`Self::get`]
     /// instead.
-    pub fn get_by_name(&self, name: &str) -> Option<Value> {
+    pub fn get_by_name(&self, name: &str) -> Result<Value, FieldAccessError> {
         if name == "label" {
             if let Some(label) = self.label() {
-                return Some(label.into_value());
+                return Ok(label.into_value());
             }
         }
-        let id = self.elem().field_id(name)?;
+        let id = self.elem().field_id(name).ok_or(FieldAccessError::Unknown)?;
         self.get(id, None)
     }
 
@@ -213,7 +226,7 @@ impl Content {
     /// generated by the `#[elem]` macro.
     pub fn field(&self, id: u8) -> StrResult<Value> {
         self.get(id, None)
-            .ok_or_else(|| missing_field(self.elem().field_name(id).unwrap()))
+            .map_err(|e| e.message(self, self.elem().field_name(id).unwrap()))
     }
 
     /// Get a field by name, returning a missing field error if it does not
@@ -222,7 +235,7 @@ impl Content {
     /// If you have access to the field IDs of the element, use [`Self::field`]
     /// instead.
     pub fn field_by_name(&self, name: &str) -> StrResult<Value> {
-        self.get_by_name(name).ok_or_else(|| missing_field(name))
+        self.get_by_name(name).map_err(|e| e.message(self, name))
     }
 
     /// Resolve all fields with the styles and save them in-place.
@@ -232,16 +245,14 @@ impl Content {
 
     /// Create a new sequence element from multiples elements.
     pub fn sequence(iter: impl IntoIterator<Item = Self>) -> Self {
-        let mut iter = iter.into_iter();
-        let Some(first) = iter.next() else { return Self::empty() };
-        let Some(second) = iter.next() else { return first };
-        SequenceElem::new(
-            std::iter::once(first)
-                .chain(std::iter::once(second))
-                .chain(iter)
-                .collect(),
-        )
-        .into()
+        let vec: Vec<_> = iter.into_iter().collect();
+        if vec.is_empty() {
+            Self::empty()
+        } else if vec.len() == 1 {
+            vec.into_iter().next().unwrap()
+        } else {
+            SequenceElem::new(vec).into()
+        }
     }
 
     /// Whether the contained element is of type `T`.
@@ -299,7 +310,7 @@ impl Content {
         // use a `*const Content` pointer.
         let vtable = self.elem().vtable()(TypeId::of::<C>())?;
         let data = self as *const Content as *const ();
-        Some(unsafe { &*fat::from_raw_parts(data, vtable) })
+        Some(unsafe { &*fat::from_raw_parts(data, vtable.as_ptr()) })
     }
 
     /// Cast to a mutable trait object if the contained element has the given
@@ -319,7 +330,7 @@ impl Content {
         // mutable access is required.
         let vtable = self.elem().vtable()(TypeId::of::<C>())?;
         let data = self as *mut Content as *mut ();
-        Some(unsafe { &mut *fat::from_raw_parts_mut(data, vtable) })
+        Some(unsafe { &mut *fat::from_raw_parts_mut(data, vtable.as_ptr()) })
     }
 
     /// Whether the content is an empty sequence.
@@ -349,7 +360,7 @@ impl Content {
         context: Tracked<Context>,
         recipe: Recipe,
     ) -> SourceResult<Self> {
-        if recipe.selector.is_none() {
+        if recipe.selector().is_none() {
             recipe.apply(engine, context, self)
         } else {
             Ok(self.styled(recipe))
@@ -385,6 +396,19 @@ impl Content {
         }
     }
 
+    /// Style this content with a full style map in-place.
+    pub fn style_in_place(&mut self, styles: Styles) {
+        if styles.is_empty() {
+            return;
+        }
+
+        if let Some(style_elem) = self.to_packed_mut::<StyledElem>() {
+            style_elem.styles.apply(styles);
+        } else {
+            *self = StyledElem::new(std::mem::take(self), styles).into();
+        }
+    }
+
     /// Queries the content tree for all elements that match the given selector.
     ///
     /// Elements produced in `show` rules will not be included in the results.
@@ -472,16 +496,7 @@ impl Content {
 
     /// Link the content somewhere.
     pub fn linked(self, dest: Destination) -> Self {
-        self.styled(MetaElem::set_data(smallvec![Meta::Link(dest)]))
-    }
-
-    /// Make the content linkable by `.linked(Destination::Location(loc))`.
-    ///
-    /// Should be used in combination with [`Location::variant`].
-    pub fn backlinked(self, loc: Location) -> Self {
-        let mut backlink = Content::empty();
-        backlink.set_location(loc);
-        self.styled(MetaElem::set_data(smallvec![Meta::Elem(backlink)]))
+        self.styled(LinkElem::set_dests(smallvec![dest]))
     }
 
     /// Set alignments for this content.
@@ -548,8 +563,8 @@ impl Content {
         default: Option<Value>,
     ) -> StrResult<Value> {
         self.get_by_name(&field)
-            .or(default)
-            .ok_or_else(|| missing_field_no_default(&field))
+            .or_else(|e| default.ok_or(e))
+            .map_err(|e| e.message_no_default(self, &field))
     }
 
     /// Returns the fields of this content.
@@ -711,7 +726,7 @@ impl<T: NativeElement> Bounds for T {
                 label: inner.label,
                 location: inner.location,
                 lifecycle: inner.lifecycle.clone(),
-                elem: LazyHash::with_hash(self.clone(), inner.elem.hash()),
+                elem: LazyHash::reuse(self.clone(), &inner.elem),
             }),
             span,
         }
@@ -758,7 +773,7 @@ impl<T: NativeElement> Packed<T> {
             // Safety:
             // - We have checked the type.
             // - Packed<T> is repr(transparent).
-            return Some(unsafe { std::mem::transmute(content) });
+            return Some(unsafe { std::mem::transmute::<&Content, &Packed<T>>(content) });
         }
         None
     }
@@ -769,7 +784,9 @@ impl<T: NativeElement> Packed<T> {
             // Safety:
             // - We have checked the type.
             // - Packed<T> is repr(transparent).
-            return Some(unsafe { std::mem::transmute(content) });
+            return Some(unsafe {
+                std::mem::transmute::<&mut Content, &mut Packed<T>>(content)
+            });
         }
         None
     }
@@ -780,7 +797,7 @@ impl<T: NativeElement> Packed<T> {
             // Safety:
             // - We have checked the type.
             // - Packed<T> is repr(transparent).
-            return Ok(unsafe { std::mem::transmute(content) });
+            return Ok(unsafe { std::mem::transmute::<Content, Packed<T>>(content) });
         }
         Err(content)
     }
@@ -901,17 +918,11 @@ impl Repr for SequenceElem {
         if self.children.is_empty() {
             "[]".into()
         } else {
-            eco_format!(
-                "[{}]",
-                crate::foundations::repr::pretty_array_like(
-                    &self
-                        .children
-                        .iter()
-                        .map(|c| c.inner.elem.repr())
-                        .collect::<Vec<_>>(),
-                    false
-                )
-            )
+            let elements = crate::foundations::repr::pretty_array_like(
+                &self.children.iter().map(|c| c.inner.elem.repr()).collect::<Vec<_>>(),
+                false,
+            );
+            eco_format!("sequence{}", elements)
         }
     }
 }
@@ -954,18 +965,43 @@ pub trait PlainText {
     fn plain_text(&self, text: &mut EcoString);
 }
 
-/// The missing field access error message.
-#[cold]
-fn missing_field(field: &str) -> EcoString {
-    eco_format!("content does not contain field {}", field.repr())
+/// An error arising when trying to access a field of content.
+#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
+pub enum FieldAccessError {
+    Unknown,
+    Unset,
+    Internal,
 }
 
-/// The missing field access error message when no default value was given.
-#[cold]
-fn missing_field_no_default(field: &str) -> EcoString {
-    eco_format!(
-        "content does not contain field {} and \
-         no default value was specified",
-        field.repr()
-    )
+impl FieldAccessError {
+    /// Formats the error message given the content and the field name.
+    #[cold]
+    pub fn message(self, content: &Content, field: &str) -> EcoString {
+        let elem_name = content.elem().name();
+        match self {
+            FieldAccessError::Unknown => {
+                eco_format!("{elem_name} does not have field {}", field.repr())
+            }
+            FieldAccessError::Unset => {
+                eco_format!(
+                    "field {} in {elem_name} is not known at this point",
+                    field.repr()
+                )
+            }
+            FieldAccessError::Internal => {
+                eco_format!(
+                    "internal error when accessing field {} in {elem_name} – this is a bug",
+                    field.repr()
+                )
+            }
+        }
+    }
+
+    /// Formats the error message for an `at` calls without a default value.
+    #[cold]
+    pub fn message_no_default(self, content: &Content, field: &str) -> EcoString {
+        let mut msg = self.message(content, field);
+        msg.push_str(" and no default was specified");
+        msg
+    }
 }
