@@ -2,7 +2,10 @@ use std::num::ParseFloatError;
 
 use ecow::{eco_format, EcoString};
 
-use crate::foundations::{cast, func, repr, scope, ty, Repr, Str};
+use crate::diag::StrResult;
+use crate::foundations::{
+    bail, cast, func, repr, scope, ty, Bytes, Decimal, Endianness, Repr, Str,
+};
 use crate::layout::Ratio;
 
 /// A floating-point number.
@@ -13,6 +16,9 @@ use crate::layout::Ratio;
 ///
 /// You can convert a value to a float with this type's constructor.
 ///
+/// NaN and positive infinity are available as `{float.nan}` and `{float.inf}`
+/// respectively.
+///
 /// # Example
 /// ```example
 /// #3.14 \
@@ -24,13 +30,22 @@ type f64;
 
 #[scope]
 impl f64 {
+    /// Positive infinity.
+    const INF: f64 = f64::INFINITY;
+
+    /// A NaN value, as defined by the
+    /// [IEEE 754 standard](https://en.wikipedia.org/wiki/IEEE_754).
+    const NAN: f64 = f64::NAN;
+
     /// Converts a value to a float.
     ///
     /// - Booleans are converted to `0.0` or `1.0`.
-    /// - Integers are converted to the closest 64-bit float.
+    /// - Integers are converted to the closest 64-bit float. For integers with
+    ///   absolute value less than `{calc.pow(2, 53)}`, this conversion is
+    ///   exact.
     /// - Ratios are divided by 100%.
-    /// - Strings are parsed in base 10 to the closest 64-bit float.
-    ///   Exponential notation is supported.
+    /// - Strings are parsed in base 10 to the closest 64-bit float. Exponential
+    ///   notation is supported.
     ///
     /// ```example
     /// #float(false) \
@@ -56,7 +71,7 @@ impl f64 {
     /// ```example
     /// #float.is-nan(0) \
     /// #float.is-nan(1) \
-    /// #float.is-nan(calc.nan)
+    /// #float.is-nan(float.nan)
     /// ```
     #[func]
     pub fn is_nan(self) -> bool {
@@ -65,13 +80,13 @@ impl f64 {
 
     /// Checks if a float is infinite.
     ///
-    /// For floats, there is positive and negative infinity. This function
-    /// returns `true` if the float is either positive or negative infinity.
+    /// Floats can represent positive infinity and negative infinity. This
+    /// function returns `{true}` if the float is an infinity.
     ///
     /// ```example
     /// #float.is-infinite(0) \
     /// #float.is-infinite(1) \
-    /// #float.is-infinite(calc.inf)
+    /// #float.is-infinite(float.inf)
     /// ```
     #[func]
     pub fn is_infinite(self) -> bool {
@@ -82,18 +97,70 @@ impl f64 {
     ///
     /// - If the number is positive (including `{+0.0}`), returns `{1.0}`.
     /// - If the number is negative (including `{-0.0}`), returns `{-1.0}`.
-    /// - If the number is [`{calc.nan}`]($calc.nan), returns
-    ///   [`{calc.nan}`]($calc.nan).
+    /// - If the number is NaN, returns `{float.nan}`.
     ///
     /// ```example
     /// #(5.0).signum() \
     /// #(-5.0).signum() \
     /// #(0.0).signum() \
+    /// #float.nan.signum()
     /// ```
     #[func]
     pub fn signum(self) -> f64 {
         f64::signum(self)
     }
+
+    /// Converts bytes to a float.
+    ///
+    /// ```example
+    /// #float.from-bytes(bytes((0, 0, 0, 0, 0, 0, 240, 63))) \
+    /// #float.from-bytes(bytes((63, 240, 0, 0, 0, 0, 0, 0)), endian: "big")
+    /// ```
+    #[func]
+    pub fn from_bytes(
+        /// The bytes that should be converted to a float.
+        ///
+        /// Must be of length exactly 8 so that the result fits into a 64-bit
+        /// float.
+        bytes: Bytes,
+        /// The endianness of the conversion.
+        #[named]
+        #[default(Endianness::Little)]
+        endian: Endianness,
+    ) -> StrResult<f64> {
+        // Convert slice to an array of length 8.
+        let buf: [u8; 8] = match bytes.as_ref().try_into() {
+            Ok(buffer) => buffer,
+            Err(_) => bail!("bytes must have a length of exactly 8"),
+        };
+
+        Ok(match endian {
+            Endianness::Little => f64::from_le_bytes(buf),
+            Endianness::Big => f64::from_be_bytes(buf),
+        })
+    }
+
+    /// Converts a float to bytes.
+    ///
+    /// ```example
+    /// #array(1.0.to-bytes(endian: "big")) \
+    /// #array(1.0.to-bytes())
+    /// ```
+    #[func]
+    pub fn to_bytes(
+        self,
+        /// The endianness of the conversion.
+        #[named]
+        #[default(Endianness::Little)]
+        endian: Endianness,
+    ) -> Bytes {
+        match endian {
+            Endianness::Little => self.to_le_bytes(),
+            Endianness::Big => self.to_be_bytes(),
+        }
+        .as_slice()
+        .into()
+    }
 }
 
 impl Repr for f64 {
@@ -110,6 +177,7 @@ cast! {
     v: f64 => Self(v),
     v: bool => Self(v as i64 as f64),
     v: i64 => Self(v as f64),
+    v: Decimal => Self(f64::try_from(v).map_err(|_| eco_format!("invalid float: {}", v))?),
     v: Ratio => Self(v.get()),
     v: Str => Self(
         parse_float(v.clone().into())
