@@ -1,7 +1,7 @@
 use ecow::{eco_format, EcoString};
 
 use crate::foundations::{func, scope, ty, Repr};
-use crate::util::PicoStr;
+use crate::utils::PicoStr;
 
 /// A label for an element.
 ///
@@ -29,6 +29,19 @@ use crate::util::PicoStr;
 /// its name in angle brackets. This works both in markup and code. A label's
 /// name can contain letters, numbers, `_`, `-`, `:`, and `.`.
 ///
+/// Note that there is a syntactical difference when using the dedicated syntax
+/// for this function. In the code below, the `[<a>]` terminates the heading and
+/// thus attaches to the heading itself, whereas the `[#label("b")]` is part of
+/// the heading and thus attaches to the heading's text.
+///
+/// ```typ
+/// // Equivalent to `#heading[Introduction] <a>`.
+/// = Introduction <a>
+///
+/// // Equivalent to `#heading[Conclusion #label("b")]`.
+/// = Conclusion #label("b")
+/// ```
+///
 /// Currently, labels can only be attached to elements in markup mode, not in
 /// code mode. This might change in the future.
 #[ty(scope, cast)]
