@@ -3,6 +3,7 @@
 use ecow::{eco_format, EcoString};
 
 use crate::foundations::{func, Str, Value};
+use crate::utils::round_with_precision;
 
 /// The Unicode minus sign.
 pub const MINUS_SIGN: &str = "\u{2212}";
@@ -73,25 +74,31 @@ pub fn format_int_with_base(mut n: i64, base: i64) -> EcoString {
 }
 
 /// Converts a float to a string representation with a specific precision and a
-/// suffix, all with a single allocation.
+/// unit, all with a single allocation.
+///
+/// The returned string is always valid Typst code. As such, it might not be a
+/// float literal. For example, it may return `"float.inf"`.
 pub fn format_float(
     mut value: f64,
     precision: Option<u8>,
     force_separator: bool,
-    suffix: &str,
+    unit: &str,
 ) -> EcoString {
     if let Some(p) = precision {
-        let offset = 10_f64.powi(p as i32);
-        value = (value * offset).round() / offset;
+        value = round_with_precision(value, p as i16);
     }
     // Debug for f64 always prints a decimal separator, while Display only does
     // when necessary.
+    let unit_multiplication = if unit.is_empty() { "" } else { " * 1" };
     if value.is_nan() {
-        "NaN".into()
+        eco_format!("float.nan{unit_multiplication}{unit}")
+    } else if value.is_infinite() {
+        let sign = if value < 0.0 { "-" } else { "" };
+        eco_format!("{sign}float.inf{unit_multiplication}{unit}")
     } else if force_separator {
-        eco_format!("{:?}{}", value, suffix)
+        eco_format!("{value:?}{unit}")
     } else {
-        eco_format!("{}{}", value, suffix)
+        eco_format!("{value}{unit}")
     }
 }
 
@@ -112,6 +119,9 @@ pub fn format_float_with_unit(value: f64, unit: &str) -> EcoString {
 pub fn display_float(value: f64) -> EcoString {
     if value.is_nan() {
         "NaN".into()
+    } else if value.is_infinite() {
+        let sign = if value < 0.0 { MINUS_SIGN } else { "" };
+        eco_format!("{sign}∞")
     } else if value < 0.0 {
         eco_format!("{}{}", MINUS_SIGN, value.abs())
     } else {
@@ -119,7 +129,7 @@ pub fn display_float(value: f64) -> EcoString {
     }
 }
 
-/// Format pieces separated with commas and a final "and" or "or".
+/// Formats pieces separated with commas and a final "and" or "or".
 pub fn separated_list(pieces: &[impl AsRef<str>], last: &str) -> String {
     let mut buf = String::new();
     for (i, part) in pieces.iter().enumerate() {
@@ -142,7 +152,7 @@ pub fn separated_list(pieces: &[impl AsRef<str>], last: &str) -> String {
     buf
 }
 
-/// Format a comma-separated list.
+/// Formats a comma-separated list.
 ///
 /// Tries to format horizontally, but falls back to vertical formatting if the
 /// pieces are too long.
@@ -173,7 +183,7 @@ pub fn pretty_comma_list(pieces: &[impl AsRef<str>], trailing_comma: bool) -> St
     buf
 }
 
-/// Format an array-like construct.
+/// Formats an array-like construct.
 ///
 /// Tries to format horizontally, but falls back to vertical formatting if the
 /// pieces are too long.
