@@ -11,11 +11,12 @@ use unicode_segmentation::UnicodeSegmentation;
 use crate::diag::{bail, At, SourceResult, StrResult};
 use crate::engine::Engine;
 use crate::foundations::{
-    cast, dict, func, repr, scope, ty, Array, Bytes, Context, Dict, Func, IntoValue,
-    Label, Repr, Type, Value, Version,
+    cast, dict, func, repr, scope, ty, Array, Bytes, Context, Decimal, Dict, Func,
+    IntoValue, Label, Repr, Type, Value, Version,
 };
 use crate::layout::Alignment;
 use crate::syntax::{Span, Spanned};
+use crate::utils::PicoStr;
 
 /// Create a new [`Str`] from a format string.
 #[macro_export]
@@ -635,7 +636,7 @@ impl Repr for EcoString {
     }
 }
 
-impl Repr for &str {
+impl Repr for str {
     fn repr(&self) -> EcoString {
         let mut r = EcoString::with_capacity(self.len() + 2);
         r.push('"');
@@ -750,6 +751,12 @@ cast! {
     v: Str => v.into(),
 }
 
+cast! {
+    PicoStr,
+    self => Value::Str(self.resolve().into()),
+    v: Str => v.as_str().into(),
+}
+
 cast! {
     String,
     self => Value::Str(self.into()),
@@ -768,6 +775,7 @@ cast! {
     ToStr,
     v: i64 => Self::Int(v),
     v: f64 => Self::Str(repr::display_float(v).into()),
+    v: Decimal => Self::Str(format_str!("{}", v)),
     v: Version => Self::Str(format_str!("{}", v)),
     v: Bytes => Self::Str(
         std::str::from_utf8(&v)
