@@ -1,32 +1,38 @@
-use std::collections::{BTreeSet, HashMap};
+use std::collections::{BTreeSet, HashMap, HashSet};
 use std::fmt::{self, Debug, Formatter};
 use std::hash::Hash;
 use std::num::NonZeroUsize;
 use std::sync::RwLock;
 
-use ecow::{eco_format, EcoVec};
-use indexmap::IndexMap;
+use ecow::EcoVec;
 use smallvec::SmallVec;
 
 use crate::diag::{bail, StrResult};
 use crate::foundations::{Content, Label, Repr, Selector};
-use crate::introspection::{Location, Meta};
+use crate::introspection::{Location, Tag};
 use crate::layout::{Frame, FrameItem, Page, Point, Position, Transform};
 use crate::model::Numbering;
-use crate::util::NonZeroExt;
+use crate::utils::NonZeroExt;
 
 /// Can be queried for elements and their positions.
-#[derive(Clone)]
+#[derive(Default, Clone)]
 pub struct Introspector {
     /// The number of pages in the document.
     pages: usize,
-    /// All introspectable elements.
-    elems: IndexMap<Location, (Content, Position)>,
-    /// Maps labels to their indices in the element list. We use a smallvec such
-    /// that if the label is unique, we don't need to allocate.
-    labels: HashMap<Label, SmallVec<[usize; 1]>>,
     /// The page numberings, indexed by page number minus 1.
     page_numberings: Vec<Option<Numbering>>,
+
+    /// All introspectable elements.
+    elems: Vec<Pair>,
+    /// Lists all elements with a specific hash key. This is used for
+    /// introspector-assisted location assignment during measurement.
+    keys: MultiMap<u128, Location>,
+
+    /// Accelerates lookup of elements by location.
+    locations: HashMap<Location, usize>,
+    /// Accelerates lookup of elements by label.
+    labels: MultiMap<Label, usize>,
+
     /// Caches queries done on the introspector. This is important because
     /// even if all top-level queries are distinct, they often have shared
     /// subqueries. Example: Individual counter queries with `before` that
@@ -34,73 +40,56 @@ pub struct Introspector {
     queries: QueryCache,
 }
 
+/// A pair of content and its position.
+type Pair = (Content, Position);
+
 impl Introspector {
-    /// Applies new frames in-place, reusing the existing allocations.
+    /// Creates an introspector for a page list.
     #[typst_macros::time(name = "introspect")]
-    pub fn rebuild(&mut self, pages: &[Page]) {
-        self.pages = pages.len();
-        self.elems.clear();
-        self.labels.clear();
-        self.page_numberings.clear();
-        self.queries.clear();
-
-        for (i, page) in pages.iter().enumerate() {
-            let page_nr = NonZeroUsize::new(1 + i).unwrap();
-            self.extract(&page.frame, page_nr, Transform::identity());
-            self.page_numberings.push(page.numbering.clone());
-        }
+    pub fn new(pages: &[Page]) -> Self {
+        IntrospectorBuilder::new().build(pages)
     }
 
-    /// Extract metadata from a frame.
-    fn extract(&mut self, frame: &Frame, page: NonZeroUsize, ts: Transform) {
-        for (pos, item) in frame.items() {
-            match item {
-                FrameItem::Group(group) => {
-                    let ts = ts
-                        .pre_concat(Transform::translate(pos.x, pos.y))
-                        .pre_concat(group.transform);
-                    self.extract(&group.frame, page, ts);
-                }
-                FrameItem::Meta(Meta::Elem(content), _)
-                    if !self.elems.contains_key(&content.location().unwrap()) =>
-                {
-                    let pos = pos.transform(ts);
-                    let ret = self.elems.insert(
-                        content.location().unwrap(),
-                        (content.clone(), Position { page, point: pos }),
-                    );
-                    assert!(ret.is_none(), "duplicate locations");
-
-                    // Build the label cache.
-                    if let Some(label) = content.label() {
-                        self.labels.entry(label).or_default().push(self.elems.len() - 1);
-                    }
-                }
-                _ => {}
-            }
-        }
-    }
-
-    /// Iterate over all locatable elements.
+    /// Iterates over all locatable elements.
     pub fn all(&self) -> impl Iterator<Item = &Content> + '_ {
-        self.elems.values().map(|(c, _)| c)
+        self.elems.iter().map(|(c, _)| c)
     }
 
-    /// Get an element by its location.
-    fn get(&self, location: &Location) -> Option<&Content> {
-        self.elems.get(location).map(|(elem, _)| elem)
+    /// Retrieves the element with the given index.
+    #[track_caller]
+    fn get_by_idx(&self, idx: usize) -> &Content {
+        &self.elems[idx].0
     }
 
-    /// Get the index of this element among all.
-    fn index(&self, elem: &Content) -> usize {
-        self.elems
-            .get_index_of(&elem.location().unwrap())
-            .unwrap_or(usize::MAX)
+    /// Retrieves the position of the element with the given index.
+    #[track_caller]
+    fn get_pos_by_idx(&self, idx: usize) -> Position {
+        self.elems[idx].1
     }
 
-    /// Perform a binary search for `elem` among the `list`.
+    /// Retrieves an element by its location.
+    fn get_by_loc(&self, location: &Location) -> Option<&Content> {
+        self.locations.get(location).map(|&idx| self.get_by_idx(idx))
+    }
+
+    /// Retrieves the position of the element with the given index.
+    fn get_pos_by_loc(&self, location: &Location) -> Option<Position> {
+        self.locations.get(location).map(|&idx| self.get_pos_by_idx(idx))
+    }
+
+    /// Performs a binary search for `elem` among the `list`.
     fn binary_search(&self, list: &[Content], elem: &Content) -> Result<usize, usize> {
-        list.binary_search_by_key(&self.index(elem), |elem| self.index(elem))
+        list.binary_search_by_key(&self.elem_index(elem), |elem| self.elem_index(elem))
+    }
+
+    /// Gets the index of this element.
+    fn elem_index(&self, elem: &Content) -> usize {
+        self.loc_index(&elem.location().unwrap())
+    }
+
+    /// Gets the index of the element with this location among all.
+    fn loc_index(&self, location: &Location) -> usize {
+        self.locations.get(location).copied().unwrap_or(usize::MAX)
     }
 }
 
@@ -108,26 +97,56 @@ impl Introspector {
 impl Introspector {
     /// Query for all matching elements.
     pub fn query(&self, selector: &Selector) -> EcoVec<Content> {
-        let hash = crate::util::hash128(selector);
+        let hash = crate::utils::hash128(selector);
         if let Some(output) = self.queries.get(hash) {
             return output;
         }
 
         let output = match selector {
-            Selector::Label(label) => self
-                .labels
-                .get(label)
-                .map(|indices| {
-                    indices.iter().map(|&index| self.elems[index].0.clone()).collect()
-                })
-                .unwrap_or_default(),
-            Selector::Elem(..) | Selector::Regex(_) | Selector::Can(_) => self
+            Selector::Elem(..) => self
                 .all()
                 .filter(|elem| selector.matches(elem, None))
                 .cloned()
                 .collect(),
             Selector::Location(location) => {
-                self.get(location).cloned().into_iter().collect()
+                self.get_by_loc(location).cloned().into_iter().collect()
+            }
+            Selector::Label(label) => self
+                .labels
+                .get(label)
+                .iter()
+                .map(|&idx| self.get_by_idx(idx).clone())
+                .collect(),
+            Selector::Or(selectors) => selectors
+                .iter()
+                .flat_map(|sel| self.query(sel))
+                .map(|elem| self.elem_index(&elem))
+                .collect::<BTreeSet<usize>>()
+                .into_iter()
+                .map(|idx| self.get_by_idx(idx).clone())
+                .collect(),
+            Selector::And(selectors) => {
+                let mut results: Vec<_> =
+                    selectors.iter().map(|sel| self.query(sel)).collect();
+
+                // Extract the smallest result list and then keep only those
+                // elements in the smallest list that are also in all other
+                // lists.
+                results
+                    .iter()
+                    .enumerate()
+                    .min_by_key(|(_, vec)| vec.len())
+                    .map(|(i, _)| i)
+                    .map(|i| results.swap_remove(i))
+                    .iter()
+                    .flatten()
+                    .filter(|candidate| {
+                        results
+                            .iter()
+                            .all(|other| self.binary_search(other, candidate).is_ok())
+                    })
+                    .cloned()
+                    .collect()
             }
             Selector::Before { selector, end, inclusive } => {
                 let mut list = self.query(selector);
@@ -157,37 +176,8 @@ impl Introspector {
                 }
                 list
             }
-            Selector::And(selectors) => {
-                let mut results: Vec<_> =
-                    selectors.iter().map(|sel| self.query(sel)).collect();
-
-                // Extract the smallest result list and then keep only those
-                // elements in the smallest list that are also in all other
-                // lists.
-                results
-                    .iter()
-                    .enumerate()
-                    .min_by_key(|(_, vec)| vec.len())
-                    .map(|(i, _)| i)
-                    .map(|i| results.swap_remove(i))
-                    .iter()
-                    .flatten()
-                    .filter(|candidate| {
-                        results
-                            .iter()
-                            .all(|other| self.binary_search(other, candidate).is_ok())
-                    })
-                    .cloned()
-                    .collect()
-            }
-            Selector::Or(selectors) => selectors
-                .iter()
-                .flat_map(|sel| self.query(sel))
-                .map(|elem| self.index(&elem))
-                .collect::<BTreeSet<usize>>()
-                .into_iter()
-                .map(|index| self.elems[index].0.clone())
-                .collect(),
+            // Not supported here.
+            Selector::Can(_) | Selector::Regex(_) => EcoVec::new(),
         };
 
         self.queries.insert(hash, output.clone());
@@ -197,7 +187,12 @@ impl Introspector {
     /// Query for the first element that matches the selector.
     pub fn query_first(&self, selector: &Selector) -> Option<Content> {
         match selector {
-            Selector::Location(location) => self.get(location).cloned(),
+            Selector::Location(location) => self.get_by_loc(location).cloned(),
+            Selector::Label(label) => self
+                .labels
+                .get(label)
+                .first()
+                .map(|&idx| self.get_by_idx(idx).clone()),
             _ => self.query(selector).first().cloned(),
         }
     }
@@ -206,7 +201,7 @@ impl Introspector {
     pub fn query_unique(&self, selector: &Selector) -> StrResult<Content> {
         match selector {
             Selector::Location(location) => self
-                .get(location)
+                .get_by_loc(location)
                 .cloned()
                 .ok_or_else(|| "element does not exist in the document".into()),
             Selector::Label(label) => self.query_label(*label).cloned(),
@@ -225,15 +220,26 @@ impl Introspector {
 
     /// Query for a unique element with the label.
     pub fn query_label(&self, label: Label) -> StrResult<&Content> {
-        let indices = self.labels.get(&label).ok_or_else(|| {
-            eco_format!("label `{}` does not exist in the document", label.repr())
-        })?;
-
-        if indices.len() > 1 {
-            bail!("label `{}` occurs multiple times in the document", label.repr());
+        match *self.labels.get(&label) {
+            [idx] => Ok(self.get_by_idx(idx)),
+            [] => bail!("label `{}` does not exist in the document", label.repr()),
+            _ => bail!("label `{}` occurs multiple times in the document", label.repr()),
         }
+    }
 
-        Ok(&self.elems[indices[0]].0)
+    /// This is an optimized version of
+    /// `query(selector.before(end, true).len()` used by counters and state.
+    pub fn query_count_before(&self, selector: &Selector, end: Location) -> usize {
+        // See `query()` for details.
+        let list = self.query(selector);
+        if let Some(end) = self.get_by_loc(&end) {
+            match self.binary_search(&list, end) {
+                Ok(i) => i + 1,
+                Err(i) => i,
+            }
+        } else {
+            list.len()
+        }
     }
 
     /// The total number pages.
@@ -241,6 +247,17 @@ impl Introspector {
         NonZeroUsize::new(self.pages).unwrap_or(NonZeroUsize::ONE)
     }
 
+    /// Find the page number for the given location.
+    pub fn page(&self, location: Location) -> NonZeroUsize {
+        self.position(location).page
+    }
+
+    /// Find the position for the given location.
+    pub fn position(&self, location: Location) -> Position {
+        self.get_pos_by_loc(&location)
+            .unwrap_or(Position { page: NonZeroUsize::ONE, point: Point::zero() })
+    }
+
     /// Gets the page numbering for the given location, if any.
     pub fn page_numbering(&self, location: Location) -> Option<&Numbering> {
         let page = self.page(location);
@@ -249,29 +266,19 @@ impl Introspector {
             .and_then(|slot| slot.as_ref())
     }
 
-    /// Find the page number for the given location.
-    pub fn page(&self, location: Location) -> NonZeroUsize {
-        self.position(location).page
-    }
-
-    /// Find the position for the given location.
-    pub fn position(&self, location: Location) -> Position {
-        self.elems
-            .get(&location)
-            .map(|(_, loc)| *loc)
-            .unwrap_or(Position { page: NonZeroUsize::ONE, point: Point::zero() })
-    }
-}
-
-impl Default for Introspector {
-    fn default() -> Self {
-        Self {
-            pages: 0,
-            elems: IndexMap::new(),
-            labels: HashMap::new(),
-            page_numberings: vec![],
-            queries: QueryCache::default(),
-        }
+    /// Try to find a location for an element with the given `key` hash
+    /// that is closest after the `anchor`.
+    ///
+    /// This is used for introspector-assisted location assignment during
+    /// measurement. See the "Dealing with Measurement" section of the
+    /// [`Locator`](crate::introspection::Locator) docs for more details.
+    pub fn locator(&self, key: u128, anchor: Location) -> Option<Location> {
+        let anchor = self.loc_index(&anchor);
+        self.keys
+            .get(&key)
+            .iter()
+            .copied()
+            .min_by_key(|loc| self.loc_index(loc).wrapping_sub(anchor))
     }
 }
 
@@ -281,6 +288,33 @@ impl Debug for Introspector {
     }
 }
 
+/// A map from one keys to multiple elements.
+#[derive(Clone)]
+struct MultiMap<K, V>(HashMap<K, SmallVec<[V; 1]>>);
+
+impl<K, V> MultiMap<K, V>
+where
+    K: Hash + Eq,
+{
+    fn get(&self, key: &K) -> &[V] {
+        self.0.get(key).map_or(&[], |vec| vec.as_slice())
+    }
+
+    fn insert(&mut self, key: K, value: V) {
+        self.0.entry(key).or_default().push(value);
+    }
+
+    fn take(&mut self, key: &K) -> Option<impl Iterator<Item = V>> {
+        self.0.remove(key).map(|vec| vec.into_iter())
+    }
+}
+
+impl<K, V> Default for MultiMap<K, V> {
+    fn default() -> Self {
+        Self(HashMap::new())
+    }
+}
+
 /// Caches queries.
 #[derive(Default)]
 struct QueryCache(RwLock<HashMap<u128, EcoVec<Content>>>);
@@ -293,10 +327,6 @@ impl QueryCache {
     fn insert(&self, hash: u128, output: EcoVec<Content>) {
         self.0.write().unwrap().insert(hash, output);
     }
-
-    fn clear(&mut self) {
-        self.0.get_mut().unwrap().clear();
-    }
 }
 
 impl Clone for QueryCache {
@@ -304,3 +334,120 @@ impl Clone for QueryCache {
         Self(RwLock::new(self.0.read().unwrap().clone()))
     }
 }
+
+/// Builds the introspector.
+#[derive(Default)]
+struct IntrospectorBuilder {
+    page_numberings: Vec<Option<Numbering>>,
+    seen: HashSet<Location>,
+    insertions: MultiMap<Location, Vec<Pair>>,
+    keys: MultiMap<u128, Location>,
+    locations: HashMap<Location, usize>,
+    labels: MultiMap<Label, usize>,
+}
+
+impl IntrospectorBuilder {
+    /// Create an empty builder.
+    fn new() -> Self {
+        Self::default()
+    }
+
+    /// Build the introspector.
+    fn build(mut self, pages: &[Page]) -> Introspector {
+        self.page_numberings.reserve(pages.len());
+
+        // Discover all elements.
+        let mut root = Vec::new();
+        for (i, page) in pages.iter().enumerate() {
+            self.page_numberings.push(page.numbering.clone());
+            self.discover(
+                &mut root,
+                &page.frame,
+                NonZeroUsize::new(1 + i).unwrap(),
+                Transform::identity(),
+            );
+        }
+
+        self.locations.reserve(self.seen.len());
+
+        // Save all pairs and their descendants in the correct order.
+        let mut elems = Vec::with_capacity(self.seen.len());
+        for pair in root {
+            self.visit(&mut elems, pair);
+        }
+
+        Introspector {
+            pages: pages.len(),
+            page_numberings: self.page_numberings,
+            elems,
+            keys: self.keys,
+            locations: self.locations,
+            labels: self.labels,
+            queries: QueryCache::default(),
+        }
+    }
+
+    /// Processes the tags in the frame.
+    fn discover(
+        &mut self,
+        sink: &mut Vec<Pair>,
+        frame: &Frame,
+        page: NonZeroUsize,
+        ts: Transform,
+    ) {
+        for (pos, item) in frame.items() {
+            match item {
+                FrameItem::Group(group) => {
+                    let ts = ts
+                        .pre_concat(Transform::translate(pos.x, pos.y))
+                        .pre_concat(group.transform);
+
+                    if let Some(parent) = group.parent {
+                        let mut nested = vec![];
+                        self.discover(&mut nested, &group.frame, page, ts);
+                        self.insertions.insert(parent, nested);
+                    } else {
+                        self.discover(sink, &group.frame, page, ts);
+                    }
+                }
+                FrameItem::Tag(Tag::Start(elem)) => {
+                    let loc = elem.location().unwrap();
+                    if self.seen.insert(loc) {
+                        let point = pos.transform(ts);
+                        sink.push((elem.clone(), Position { page, point }));
+                    }
+                }
+                FrameItem::Tag(Tag::End(loc, key)) => {
+                    self.keys.insert(*key, *loc);
+                }
+                _ => {}
+            }
+        }
+    }
+
+    /// Saves a pair and all its descendants into `elems` and populates the
+    /// acceleration structures.
+    fn visit(&mut self, elems: &mut Vec<Pair>, pair: Pair) {
+        let elem = &pair.0;
+        let loc = elem.location().unwrap();
+        let idx = elems.len();
+
+        // Populate the location acceleration map.
+        self.locations.insert(loc, idx);
+
+        // Populate the label acceleration map.
+        if let Some(label) = elem.label() {
+            self.labels.insert(label, idx);
+        }
+
+        // Save the element.
+        elems.push(pair);
+
+        // Process potential descendants.
+        if let Some(insertions) = self.insertions.take(&loc) {
+            for pair in insertions.flatten() {
+                self.visit(elems, pair);
+            }
+        }
+    }
+}
