@@ -4,8 +4,8 @@ use std::ops::{BitAnd, BitAndAssign, BitOr, BitOrAssign, Deref, Not};
 
 use crate::diag::bail;
 use crate::foundations::{array, cast, Array, Resolve, Smart, StyleChain};
-use crate::layout::{Abs, Dir, Length, Ratio, Rel};
-use crate::util::Get;
+use crate::layout::{Abs, Dir, Length, Ratio, Rel, Size};
+use crate::utils::Get;
 
 /// A container with a horizontal and vertical component.
 #[derive(Default, Copy, Clone, Eq, PartialEq, Hash)]
@@ -120,6 +120,16 @@ impl<T: Ord> Axes<T> {
     }
 }
 
+impl Axes<Rel<Abs>> {
+    /// Evaluate the axes relative to the given `size`.
+    pub fn relative_to(&self, size: Size) -> Size {
+        Size {
+            x: self.x.relative_to(size.x),
+            y: self.y.relative_to(size.y),
+        }
+    }
+}
+
 impl<T> Get<Axis> for Axes<T> {
     type Component = T;
 
