@@ -1,24 +1,24 @@
 use std::borrow::Cow;
 use std::fmt::{self, Debug, Formatter};
-use std::ops::Range;
 use std::str::FromStr;
 use std::sync::Arc;
 
 use az::SaturatingAs;
 use ecow::EcoString;
-use rustybuzz::{ShapePlan, Tag, UnicodeBuffer};
+use rustybuzz::{BufferFlags, ShapePlan, UnicodeBuffer};
+use ttf_parser::Tag;
+use unicode_bidi::{BidiInfo, Level as BidiLevel};
 use unicode_script::{Script, UnicodeScript};
 
-use super::SpanMapper;
+use super::{Item, Range, SpanMapper};
 use crate::engine::Engine;
-use crate::foundations::StyleChain;
+use crate::foundations::{Smart, StyleChain};
 use crate::layout::{Abs, Dir, Em, Frame, FrameItem, Point, Size};
-use crate::syntax::Span;
 use crate::text::{
-    decorate, families, features, variant, Font, FontVariant, Glyph, Lang, Region,
-    TextElem, TextItem,
+    decorate, families, features, is_default_ignorable, variant, Font, FontVariant,
+    Glyph, Lang, Region, TextEdgeBounds, TextElem, TextItem,
 };
-use crate::util::SliceExt;
+use crate::utils::SliceExt;
 use crate::World;
 
 /// The result of shaping text.
@@ -26,7 +26,8 @@ use crate::World;
 /// This type contains owned or borrowed shaped text runs, which can be
 /// measured, used to reshape substrings more quickly and converted into a
 /// frame.
-pub(super) struct ShapedText<'a> {
+#[derive(Clone)]
+pub struct ShapedText<'a> {
     /// The start of the text in the full paragraph.
     pub base: usize,
     /// The text that was shaped.
@@ -51,7 +52,7 @@ pub(super) struct ShapedText<'a> {
 
 /// A single glyph resulting from shaping.
 #[derive(Debug, Clone)]
-pub(super) struct ShapedGlyph {
+pub struct ShapedGlyph {
     /// The font the glyph is contained in.
     pub font: Font,
     /// The glyph's index in the font.
@@ -68,20 +69,17 @@ pub(super) struct ShapedGlyph {
     /// is a sequence of one or multiple glyphs that cannot be separated and
     /// must always be treated as a union.
     ///
-    /// The range values of the glyphs in a [`ShapedText`] should not
-    /// overlap with each other, and they should be monotonically
-    /// increasing (for left-to-right or top-to-bottom text) or
-    /// monotonically decreasing (for right-to-left or bottom-to-top
-    /// text).
-    pub range: Range<usize>,
+    /// The range values of the glyphs in a [`ShapedText`] should not overlap
+    /// with each other, and they should be monotonically increasing (for
+    /// left-to-right or top-to-bottom text) or monotonically decreasing (for
+    /// right-to-left or bottom-to-top text).
+    pub range: Range,
     /// Whether splitting the shaping result before this glyph would yield the
     /// same results as shaping the parts to both sides of `text_index`
     /// separately.
     pub safe_to_break: bool,
     /// The first char in this glyph's cluster.
     pub c: char,
-    /// The source code location of the glyph and its byte offset within it.
-    pub span: (Span, u16),
     /// Whether this glyph is justifiable for CJK scripts.
     pub is_justifiable: bool,
     /// The script of the glyph.
@@ -89,8 +87,8 @@ pub(super) struct ShapedGlyph {
 }
 
 #[derive(Debug, Clone, Default)]
-pub(super) struct Adjustability {
-    /// The left and right strechability
+pub struct Adjustability {
+    /// The left and right stretchability
     pub stretchability: (Em, Em),
     /// The left and right shrinkability
     pub shrinkability: (Em, Em),
@@ -214,6 +212,7 @@ impl<'a> ShapedText<'a> {
     pub fn build(
         &self,
         engine: &Engine,
+        spans: &SpanMapper,
         justification_ratio: f64,
         extra_justification: Abs,
     ) -> Frame {
@@ -225,7 +224,6 @@ impl<'a> ShapedText<'a> {
         frame.set_baseline(top);
 
         let shift = TextElem::baseline_in(self.styles);
-        let lang = TextElem::lang_in(self.styles);
         let decos = TextElem::deco_in(self.styles);
         let fill = TextElem::fill_in(self.styles);
         let stroke = TextElem::stroke_in(self.styles);
@@ -269,7 +267,7 @@ impl<'a> ShapedText<'a> {
                     // We may not be able to reach the offset completely if
                     // it exceeds u16, but better to have a roughly correct
                     // span offset than nothing.
-                    let mut span = shaped.span;
+                    let mut span = spans.span_at(shaped.range.start);
                     span.1 = span.1.saturating_add(span_offset.saturating_as());
 
                     // |<---- a Glyph ---->|
@@ -306,7 +304,8 @@ impl<'a> ShapedText<'a> {
             let item = TextItem {
                 font,
                 size: self.size,
-                lang,
+                lang: self.lang,
+                region: self.region,
                 fill: fill.clone(),
                 stroke: stroke.clone().map(|s| s.unwrap_or_default()),
                 text: self.text[range.start - self.base..range.end - self.base].into(),
@@ -331,7 +330,7 @@ impl<'a> ShapedText<'a> {
     }
 
     /// Measure the top and bottom extent of this text.
-    fn measure(&self, engine: &Engine) -> (Abs, Abs) {
+    pub fn measure(&self, engine: &Engine) -> (Abs, Abs) {
         let mut top = Abs::zero();
         let mut bottom = Abs::zero();
 
@@ -339,9 +338,10 @@ impl<'a> ShapedText<'a> {
         let bottom_edge = TextElem::bottom_edge_in(self.styles);
 
         // Expand top and bottom by reading the font's vertical metrics.
-        let mut expand = |font: &Font, bbox: Option<ttf_parser::Rect>| {
-            top.set_max(top_edge.resolve(self.size, font, bbox));
-            bottom.set_max(-bottom_edge.resolve(self.size, font, bbox));
+        let mut expand = |font: &Font, bounds: TextEdgeBounds| {
+            let (t, b) = font.edges(top_edge, bottom_edge, self.size, bounds);
+            top.set_max(t);
+            bottom.set_max(b);
         };
 
         if self.glyphs.is_empty() {
@@ -354,18 +354,13 @@ impl<'a> ShapedText<'a> {
                     .select(family, self.variant)
                     .and_then(|id| world.font(id))
                 {
-                    expand(&font, None);
+                    expand(&font, TextEdgeBounds::Zero);
                     break;
                 }
             }
         } else {
             for g in self.glyphs.iter() {
-                let bbox = if top_edge.is_bounds() || bottom_edge.is_bounds() {
-                    g.font.ttf().glyph_bounding_box(ttf_parser::GlyphId(g.glyph_id))
-                } else {
-                    None
-                };
-                expand(&g.font, bbox);
+                expand(&g.font, TextEdgeBounds::Glyph(g.glyph_id));
             }
         }
 
@@ -409,12 +404,7 @@ impl<'a> ShapedText<'a> {
     /// shaping process if possible.
     ///
     /// The text `range` is relative to the whole paragraph.
-    pub fn reshape(
-        &'a self,
-        engine: &Engine,
-        spans: &SpanMapper,
-        text_range: Range<usize>,
-    ) -> ShapedText<'a> {
+    pub fn reshape(&'a self, engine: &Engine, text_range: Range) -> ShapedText<'a> {
         let text = &self.text[text_range.start - self.base..text_range.end - self.base];
         if let Some(glyphs) = self.slice_safe_to_break(text_range.clone()) {
             #[cfg(debug_assertions)]
@@ -436,7 +426,6 @@ impl<'a> ShapedText<'a> {
                 engine,
                 text_range.start,
                 text,
-                spans,
                 self.styles,
                 self.dir,
                 self.lang,
@@ -445,8 +434,27 @@ impl<'a> ShapedText<'a> {
         }
     }
 
+    /// Derive an empty text run with the same properties as this one.
+    pub fn empty(&self) -> Self {
+        Self {
+            text: "",
+            width: Abs::zero(),
+            glyphs: Cow::Borrowed(&[]),
+            ..*self
+        }
+    }
+
     /// Push a hyphen to end of the text.
     pub fn push_hyphen(&mut self, engine: &Engine, fallback: bool) {
+        self.insert_hyphen(engine, fallback, Side::Right)
+    }
+
+    /// Prepend a hyphen to start of the text.
+    pub fn prepend_hyphen(&mut self, engine: &Engine, fallback: bool) {
+        self.insert_hyphen(engine, fallback, Side::Left)
+    }
+
+    fn insert_hyphen(&mut self, engine: &Engine, fallback: bool, side: Side) {
         let world = engine.world;
         let book = world.book();
         let fallback_func = if fallback {
@@ -464,17 +472,17 @@ impl<'a> ShapedText<'a> {
             let ttf = font.ttf();
             let glyph_id = ttf.glyph_index('-')?;
             let x_advance = font.to_em(ttf.glyph_hor_advance(glyph_id)?);
-            let range = self
-                .glyphs
-                .last()
-                .map(|g| g.range.end..g.range.end)
-                // In the unlikely chance that we hyphenate after an empty line,
-                // ensure that the glyph range still falls after self.base so
-                // that subtracting either of the endpoints by self.base doesn't
-                // underflow. See <https://github.com/typst/typst/issues/2283>.
-                .unwrap_or_else(|| self.base..self.base);
+            let range = match side {
+                Side::Left => self.glyphs.first().map(|g| g.range.start..g.range.start),
+                Side::Right => self.glyphs.last().map(|g| g.range.end..g.range.end),
+            }
+            // In the unlikely chance that we hyphenate after an empty line,
+            // ensure that the glyph range still falls after self.base so
+            // that subtracting either of the endpoints by self.base doesn't
+            // underflow. See <https://github.com/typst/typst/issues/2283>.
+            .unwrap_or_else(|| self.base..self.base);
             self.width += x_advance.at(self.size);
-            self.glyphs.to_mut().push(ShapedGlyph {
+            let glyph = ShapedGlyph {
                 font,
                 glyph_id: glyph_id.0,
                 x_advance,
@@ -484,30 +492,33 @@ impl<'a> ShapedText<'a> {
                 range,
                 safe_to_break: true,
                 c: '-',
-                span: (Span::detached(), 0),
                 is_justifiable: false,
                 script: Script::Common,
-            });
+            };
+            match side {
+                Side::Left => self.glyphs.to_mut().insert(0, glyph),
+                Side::Right => self.glyphs.to_mut().push(glyph),
+            }
             Some(())
         });
     }
 
     /// Find the subslice of glyphs that represent the given text range if both
     /// sides are safe to break.
-    fn slice_safe_to_break(&self, text_range: Range<usize>) -> Option<&[ShapedGlyph]> {
+    fn slice_safe_to_break(&self, text_range: Range) -> Option<&[ShapedGlyph]> {
         let Range { mut start, mut end } = text_range;
         if !self.dir.is_positive() {
             std::mem::swap(&mut start, &mut end);
         }
 
-        let left = self.find_safe_to_break(start, Side::Left)?;
-        let right = self.find_safe_to_break(end, Side::Right)?;
+        let left = self.find_safe_to_break(start)?;
+        let right = self.find_safe_to_break(end)?;
         Some(&self.glyphs[left..right])
     }
 
     /// Find the glyph offset matching the text index that is most towards the
-    /// given side and safe-to-break.
-    fn find_safe_to_break(&self, text_index: usize, towards: Side) -> Option<usize> {
+    /// start of the text and safe-to-break.
+    fn find_safe_to_break(&self, text_index: usize) -> Option<usize> {
         let ltr = self.dir.is_positive();
 
         // Handle edge cases.
@@ -527,6 +538,7 @@ impl<'a> ShapedText<'a> {
                 ordering.reverse()
             }
         });
+
         let mut idx = match found {
             Ok(idx) => idx,
             Err(idx) => {
@@ -538,8 +550,9 @@ impl<'a> ShapedText<'a> {
                 // text_index:    ^
                 // glyphs:  0     .  1
                 //
-                // We will get found = Err(1), because '\n' does not have a glyph.
-                // But it's safe to break here. Thus the following condition:
+                // We will get found = Err(1), because '\n' does not have a
+                // glyph. But it's safe to break here. Thus the following
+                // condition:
                 // - glyphs[0].end == text_index == 3
                 // - text[3] == '\n'
                 return (idx > 0
@@ -549,13 +562,11 @@ impl<'a> ShapedText<'a> {
             }
         };
 
-        let next = match towards {
-            Side::Left => usize::checked_sub,
-            Side::Right => usize::checked_add,
-        };
-
-        // Search for the outermost glyph with the text index.
-        while let Some(next) = next(idx, 1) {
+        // Search for the start-most glyph with the text index. This means
+        // we take empty range glyphs at the start and leave those at the end
+        // for the next line.
+        let dec = if ltr { usize::checked_sub } else { usize::checked_add };
+        while let Some(next) = dec(idx, 1) {
             if self.glyphs.get(next).map_or(true, |g| g.range.start != text_index) {
                 break;
             }
@@ -575,27 +586,77 @@ impl Debug for ShapedText<'_> {
     }
 }
 
-/// Holds shaping results and metadata common to all shaped segments.
-struct ShapingContext<'a, 'v> {
-    engine: &'a Engine<'v>,
-    spans: &'a SpanMapper,
-    glyphs: Vec<ShapedGlyph>,
-    used: Vec<Font>,
+/// Group a range of text by BiDi level and script, shape the runs and generate
+/// items for them.
+pub fn shape_range<'a>(
+    items: &mut Vec<(Range, Item<'a>)>,
+    engine: &Engine,
+    text: &'a str,
+    bidi: &BidiInfo<'a>,
+    range: Range,
     styles: StyleChain<'a>,
-    size: Abs,
-    variant: FontVariant,
-    features: Vec<rustybuzz::Feature>,
-    fallback: bool,
-    dir: Dir,
+) {
+    let script = TextElem::script_in(styles);
+    let lang = TextElem::lang_in(styles);
+    let region = TextElem::region_in(styles);
+    let mut process = |range: Range, level: BidiLevel| {
+        let dir = if level.is_ltr() { Dir::LTR } else { Dir::RTL };
+        let shaped =
+            shape(engine, range.start, &text[range.clone()], styles, dir, lang, region);
+        items.push((range, Item::Text(shaped)));
+    };
+
+    let mut prev_level = BidiLevel::ltr();
+    let mut prev_script = Script::Unknown;
+    let mut cursor = range.start;
+
+    // Group by embedding level and script.  If the text's script is explicitly
+    // set (rather than inferred from the glyphs), we keep the script at an
+    // unchanging `Script::Unknown` so that only level changes cause breaks.
+    for i in range.clone() {
+        if !text.is_char_boundary(i) {
+            continue;
+        }
+
+        let level = bidi.levels[i];
+        let curr_script = match script {
+            Smart::Auto => {
+                text[i..].chars().next().map_or(Script::Unknown, |c| c.script())
+            }
+            Smart::Custom(_) => Script::Unknown,
+        };
+
+        if level != prev_level || !is_compatible(curr_script, prev_script) {
+            if cursor < i {
+                process(cursor..i, prev_level);
+            }
+            cursor = i;
+            prev_level = level;
+            prev_script = curr_script;
+        } else if is_generic_script(prev_script) {
+            prev_script = curr_script;
+        }
+    }
+
+    process(cursor..range.end, prev_level);
+}
+
+/// Whether this is not a specific script.
+fn is_generic_script(script: Script) -> bool {
+    matches!(script, Script::Unknown | Script::Common | Script::Inherited)
+}
+
+/// Whether these script can be part of the same shape run.
+fn is_compatible(a: Script, b: Script) -> bool {
+    is_generic_script(a) || is_generic_script(b) || a == b
 }
 
 /// Shape text into [`ShapedText`].
 #[allow(clippy::too_many_arguments)]
-pub(super) fn shape<'a>(
+fn shape<'a>(
     engine: &Engine,
     base: usize,
     text: &'a str,
-    spans: &SpanMapper,
     styles: StyleChain<'a>,
     dir: Dir,
     lang: Lang,
@@ -604,7 +665,6 @@ pub(super) fn shape<'a>(
     let size = TextElem::size_in(styles);
     let mut ctx = ShapingContext {
         engine,
-        spans,
         size,
         glyphs: vec![],
         used: vec![],
@@ -641,6 +701,19 @@ pub(super) fn shape<'a>(
     }
 }
 
+/// Holds shaping results and metadata common to all shaped segments.
+struct ShapingContext<'a, 'v> {
+    engine: &'a Engine<'v>,
+    glyphs: Vec<ShapedGlyph>,
+    used: Vec<Font>,
+    styles: StyleChain<'a>,
+    size: Abs,
+    variant: FontVariant,
+    features: Vec<rustybuzz::Feature>,
+    fallback: bool,
+    dir: Dir,
+}
+
 /// Shape text with font fallback using the `families` iterator.
 fn shape_segment<'a>(
     ctx: &mut ShapingContext,
@@ -648,8 +721,11 @@ fn shape_segment<'a>(
     text: &str,
     mut families: impl Iterator<Item = &'a str> + Clone,
 ) {
-    // Fonts dont have newlines and tabs.
-    if text.chars().all(|c| c == '\n' || c == '\t') {
+    // Don't try shaping newlines, tabs, or default ignorables.
+    if text
+        .chars()
+        .all(|c| c == '\n' || c == '\t' || is_default_ignorable(c))
+    {
         return;
     }
 
@@ -697,9 +773,14 @@ fn shape_segment<'a>(
     });
     buffer.guess_segment_properties();
 
+    // By default, Harfbuzz will create zero-width space glyphs for default
+    // ignorables. This is probably useful for GUI apps that want noticable
+    // effects on the cursor for those, but for us it's not useful and hurts
+    // text extraction.
+    buffer.set_flags(BufferFlags::REMOVE_DEFAULT_IGNORABLES);
+
     // Prepare the shape plan. This plan depends on direction, script, language,
-    // and features, but is independent from the text and can thus be
-    // memoized.
+    // and features, but is independent from the text and can thus be memoized.
     let plan = create_shape_plan(
         &font,
         buffer.direction(),
@@ -744,7 +825,6 @@ fn shape_segment<'a>(
                 range: start..end,
                 safe_to_break: !info.unsafe_to_break(),
                 c,
-                span: ctx.spans.span_at(start),
                 is_justifiable: is_justifiable(
                     c,
                     script,
@@ -835,7 +915,6 @@ fn shape_tofus(ctx: &mut ShapingContext, base: usize, text: &str, font: Font) {
             range: start..end,
             safe_to_break: true,
             c,
-            span: ctx.spans.span_at(start),
             is_justifiable: is_justifiable(
                 c,
                 script,
@@ -894,7 +973,7 @@ fn calculate_adjustability(ctx: &mut ShapingContext, lang: Lang, region: Option<
             continue;
         }
 
-        // Now we apply consecutive punctuation adjustment, specified in Chinese Layout
+        // Now we apply consecutive punctuation adjustment, specified in Chinese Layout.
         // Requirements, section 3.1.6.1 Punctuation Adjustment Space, and Japanese Layout
         // Requirements, section 3.1 Line Composition Rules for Punctuation Marks
         let Some(next) = glyphs.peek_mut() else { continue };
@@ -931,7 +1010,7 @@ fn language(styles: StyleChain) -> rustybuzz::Language {
 
 /// Returns true if all glyphs in `glyphs` have ranges within the range `range`.
 #[cfg(debug_assertions)]
-fn assert_all_glyphs_in_range(glyphs: &[ShapedGlyph], text: &str, range: Range<usize>) {
+fn assert_all_glyphs_in_range(glyphs: &[ShapedGlyph], text: &str, range: Range) {
     if glyphs
         .iter()
         .any(|g| g.range.start < range.start || g.range.end > range.end)
@@ -940,9 +1019,11 @@ fn assert_all_glyphs_in_range(glyphs: &[ShapedGlyph], text: &str, range: Range<u
     }
 }
 
-/// Asserts that the ranges of `glyphs` is in the proper order according to `dir`.
+/// Asserts that the ranges of `glyphs` is in the proper order according to
+/// `dir`.
 ///
-/// This asserts instead of returning a bool in order to provide a more informative message when the invariant is violated.
+/// This asserts instead of returning a bool in order to provide a more
+/// informative message when the invariant is violated.
 #[cfg(debug_assertions)]
 fn assert_glyph_ranges_in_order(glyphs: &[ShapedGlyph], dir: Dir) {
     if glyphs.is_empty() {
@@ -967,15 +1048,15 @@ fn assert_glyph_ranges_in_order(glyphs: &[ShapedGlyph], dir: Dir) {
 }
 
 // The CJK punctuation that can appear at the beginning or end of a line.
-pub(super) const BEGIN_PUNCT_PAT: &[char] =
+pub const BEGIN_PUNCT_PAT: &[char] =
     &['“', '‘', '《', '〈', '（', '『', '「', '【', '〖', '〔', '［', '｛'];
-pub(super) const END_PUNCT_PAT: &[char] = &[
+pub const END_PUNCT_PAT: &[char] = &[
     '”', '’', '，', '．', '。', '、', '：', '；', '》', '〉', '）', '』', '」', '】',
     '〗', '〕', '］', '｝', '？', '！',
 ];
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
-pub(super) enum CjkPunctStyle {
+pub enum CjkPunctStyle {
     /// Standard GB/T 15834-2011, used mostly in mainland China.
     Gb,
     /// Standard by Taiwan Ministry of Education, used in Taiwan and Hong Kong.
@@ -984,7 +1065,7 @@ pub(super) enum CjkPunctStyle {
     Jis,
 }
 
-pub(super) fn cjk_punct_style(lang: Lang, region: Option<Region>) -> CjkPunctStyle {
+pub fn cjk_punct_style(lang: Lang, region: Option<Region>) -> CjkPunctStyle {
     match (lang, region.as_ref().map(Region::as_str)) {
         (Lang::CHINESE, Some("TW" | "HK")) => CjkPunctStyle::Cns,
         (Lang::JAPANESE, _) => CjkPunctStyle::Jis,
@@ -999,7 +1080,7 @@ fn is_space(c: char) -> bool {
 }
 
 /// Whether the glyph is part of Chinese or Japanese script (i.e. CJ, not CJK).
-pub(super) fn is_of_cj_script(c: char) -> bool {
+pub fn is_of_cj_script(c: char) -> bool {
     is_cj_script(c, c.script())
 }
 
@@ -1033,8 +1114,9 @@ fn is_cjk_left_aligned_punctuation(
     }
 
     if matches!(style, Gb) && matches!(c, '？' | '！') {
-        // In GB style, exclamations and question marks are also left aligned and can be adjusted.
-        // Note that they are not adjustable in other styles.
+        // In GB style, exclamations and question marks are also left aligned
+        // and can be adjusted. Note that they are not adjustable in other
+        // styles.
         return true;
     }
 
