@@ -5,11 +5,11 @@ use std::ops::{Add, Div, Mul, Neg};
 use comemo::Tracked;
 use ecow::{eco_format, EcoString};
 
-use crate::diag::{At, Hint, HintedStrResult, SourceResult};
+use crate::diag::{bail, HintedStrResult, SourceResult};
 use crate::foundations::{func, scope, ty, Context, Fold, Repr, Resolve, StyleChain};
 use crate::layout::{Abs, Em};
 use crate::syntax::Span;
-use crate::util::Numeric;
+use crate::utils::Numeric;
 
 /// A size or distance, possibly expressed with contextual units.
 ///
@@ -81,12 +81,15 @@ impl Length {
         if self.em == Em::zero() {
             return Ok(());
         }
-        Err(eco_format!(
+
+        bail!(
+            span,
             "cannot convert a length with non-zero em units (`{}`) to {unit}",
-            self.repr()
-        ))
-        .hint(eco_format!("use `length.abs.{unit}()` instead to ignore its em component"))
-        .at(span)
+            self.repr();
+            hint: "use `length.to-absolute()` to resolve its em component \
+                   (requires context)";
+            hint: "or use `length.abs.{unit}()` instead to ignore its em component"
+        )
     }
 }
 
@@ -227,7 +230,7 @@ impl Add for Length {
     }
 }
 
-sub_impl!(Length - Length -> Length);
+typst_utils::sub_impl!(Length - Length -> Length);
 
 impl Mul<f64> for Length {
     type Output = Self;
@@ -253,10 +256,10 @@ impl Div<f64> for Length {
     }
 }
 
-assign_impl!(Length += Length);
-assign_impl!(Length -= Length);
-assign_impl!(Length *= f64);
-assign_impl!(Length /= f64);
+typst_utils::assign_impl!(Length += Length);
+typst_utils::assign_impl!(Length -= Length);
+typst_utils::assign_impl!(Length *= f64);
+typst_utils::assign_impl!(Length /= f64);
 
 impl Resolve for Length {
     type Output = Abs;
