@@ -1,6 +1,6 @@
 use std::borrow::Cow;
 use std::num::NonZeroUsize;
-use std::ptr;
+use std::ops::RangeInclusive;
 use std::str::FromStr;
 
 use comemo::Track;
@@ -8,20 +8,16 @@ use comemo::Track;
 use crate::diag::{bail, SourceResult};
 use crate::engine::Engine;
 use crate::foundations::{
-    cast, elem, AutoValue, Cast, Content, Context, Dict, Fold, Func, NativeElement,
-    Packed, Resolve, Smart, StyleChain, Value,
+    cast, elem, Args, AutoValue, Cast, Construct, Content, Context, Dict, Fold, Func,
+    NativeElement, Set, Smart, StyleChain, Value,
 };
-use crate::introspection::{Counter, CounterDisplayElem, CounterKey, ManualPageCounter};
 use crate::layout::{
-    Abs, AlignElem, Alignment, Axes, ColumnsElem, Dir, Frame, HAlignment, LayoutMultiple,
-    Length, OuterVAlignment, Point, Ratio, Regions, Rel, Sides, Size, SpecificAlignment,
-    VAlignment,
+    Abs, Alignment, FlushElem, Frame, HAlignment, Length, OuterVAlignment, Ratio, Rel,
+    Sides, SpecificAlignment,
 };
-
 use crate::model::Numbering;
-use crate::text::TextElem;
-use crate::util::{NonZeroExt, Numeric, Scalar};
-use crate::visualize::Paint;
+use crate::utils::{singleton, NonZeroExt, Scalar};
+use crate::visualize::{Color, Paint};
 
 /// Layouts its child onto one or multiple pages.
 ///
@@ -42,9 +38,12 @@ use crate::visualize::Paint;
 ///
 /// There you go, US friends!
 /// ```
-#[elem]
+#[elem(Construct)]
 pub struct PageElem {
     /// A standard paper size to set width and height.
+    ///
+    /// This is just a shorthand for setting `width` and `height` and, as such,
+    /// cannot be retrieved in a context expression.
     #[external]
     #[default(Paper::A4)]
     pub paper: Paper,
@@ -68,6 +67,7 @@ pub struct PageElem {
             .or_else(|| paper.map(|paper| Smart::Custom(paper.width().into())))
     )]
     #[default(Smart::Custom(Paper::A4.width().into()))]
+    #[ghost]
     pub width: Smart<Length>,
 
     /// The height of the page.
@@ -82,6 +82,7 @@ pub struct PageElem {
             .or_else(|| paper.map(|paper| Smart::Custom(paper.height().into())))
     )]
     #[default(Smart::Custom(Paper::A4.height().into()))]
+    #[ghost]
     pub height: Smart<Length>,
 
     /// Whether the page is flipped into landscape orientation.
@@ -103,6 +104,7 @@ pub struct PageElem {
     /// +1 555 555 5555
     /// ```
     #[default(false)]
+    #[ghost]
     pub flipped: bool,
 
     /// The page's margins.
@@ -142,6 +144,7 @@ pub struct PageElem {
     /// )
     /// ```
     #[fold]
+    #[ghost]
     pub margin: Margin,
 
     /// On which side the pages will be bound.
@@ -153,6 +156,7 @@ pub struct PageElem {
     ///
     /// This affects the meaning of the `inside` and `outside` options for
     /// margins.
+    #[ghost]
     pub binding: Smart<Binding>,
 
     /// How many columns the page has.
@@ -173,14 +177,23 @@ pub struct PageElem {
     /// of a rapidly changing climate.
     /// ```
     #[default(NonZeroUsize::ONE)]
+    #[ghost]
     pub columns: NonZeroUsize,
 
-    /// The page's background color.
+    /// The page's background fill.
     ///
-    /// This instructs the printer to color the complete page with the given
-    /// color. If you are considering larger production runs, it may be more
-    /// environmentally friendly and cost-effective to source pre-dyed pages and
-    /// not set this property.
+    /// Setting this to something non-transparent instructs the printer to color
+    /// the complete page. If you are considering larger production runs, it may
+    /// be more environmentally friendly and cost-effective to source pre-dyed
+    /// pages and not set this property.
+    ///
+    /// When set to `{none}`, the background becomes transparent. Note that PDF
+    /// pages will still appear with a (usually white) background in viewers,
+    /// but they are actually transparent. (If you print them, no color is used
+    /// for the background.)
+    ///
+    /// The default of `{auto}` results in `{none}` for PDF output, and
+    /// `{white}` for PNG and SVG.
     ///
     /// ```example
     /// #set page(fill: rgb("444352"))
@@ -188,7 +201,8 @@ pub struct PageElem {
     /// *Dark mode enabled.*
     /// ```
     #[borrowed]
-    pub fill: Option<Paint>,
+    #[ghost]
+    pub fill: Smart<Option<Paint>>,
 
     /// How to [number]($numbering) the pages.
     ///
@@ -205,6 +219,7 @@ pub struct PageElem {
     /// #lorem(48)
     /// ```
     #[borrowed]
+    #[ghost]
     pub numbering: Option<Numbering>,
 
     /// The alignment of the page numbering.
@@ -224,10 +239,16 @@ pub struct PageElem {
     /// #lorem(30)
     /// ```
     #[default(SpecificAlignment::Both(HAlignment::Center, OuterVAlignment::Bottom))]
+    #[ghost]
     pub number_align: SpecificAlignment<HAlignment, OuterVAlignment>,
 
     /// The page's header. Fills the top margin of each page.
     ///
+    /// - Content: Shows the content as the header.
+    /// - `{auto}`: Shows the page number if a `numbering` is set and
+    ///   `number-align` is `top`.
+    /// - `{none}`: Suppresses the header.
+    ///
     /// ```example
     /// #set par(justify: true)
     /// #set page(
@@ -242,17 +263,24 @@ pub struct PageElem {
     /// #lorem(19)
     /// ```
     #[borrowed]
-    pub header: Option<Content>,
+    #[ghost]
+    pub header: Smart<Option<Content>>,
 
     /// The amount the header is raised into the top margin.
     #[resolve]
     #[default(Ratio::new(0.3).into())]
+    #[ghost]
     pub header_ascent: Rel<Length>,
 
     /// The page's footer. Fills the bottom margin of each page.
     ///
-    /// For just a page number, the `numbering` property, typically suffices. If
-    /// you want to create a custom footer, but still display the page number,
+    /// - Content: Shows the content as the footer.
+    /// - `{auto}`: Shows the page number if a `numbering` is set and
+    ///   `number-align` is `bottom`.
+    /// - `{none}`: Suppresses the footer.
+    ///
+    /// For just a page number, the `numbering` property typically suffices. If
+    /// you want to create a custom footer but still display the page number,
     /// you can directly access the [page counter]($counter).
     ///
     /// ```example
@@ -273,11 +301,13 @@ pub struct PageElem {
     /// #lorem(48)
     /// ```
     #[borrowed]
-    pub footer: Option<Content>,
+    #[ghost]
+    pub footer: Smart<Option<Content>>,
 
     /// The amount the footer is lowered into the bottom margin.
     #[resolve]
     #[default(Ratio::new(0.3).into())]
+    #[ghost]
     pub footer_descent: Rel<Length>,
 
     /// Content in the page's background.
@@ -297,6 +327,7 @@ pub struct PageElem {
     /// over the world (of typesetting).
     /// ```
     #[borrowed]
+    #[ghost]
     pub background: Option<Content>,
 
     /// Content in the page's foreground.
@@ -311,6 +342,7 @@ pub struct PageElem {
     /// not understand our approach...
     /// ```
     #[borrowed]
+    #[ghost]
     pub foreground: Option<Content>,
 
     /// The contents of the page(s).
@@ -318,200 +350,86 @@ pub struct PageElem {
     /// Multiple pages will be created if the content does not fit on a single
     /// page. A new page with the page properties prior to the function invocation
     /// will be created after the body has been typeset.
+    #[external]
     #[required]
     pub body: Content,
-
-    /// Whether the page should be aligned to an even or odd page.
-    #[internal]
-    #[synthesized]
-    pub clear_to: Option<Parity>,
 }
 
-impl Packed<PageElem> {
-    /// A document can consist of multiple `PageElem`s, one per run of pages
-    /// with equal properties (not one per actual output page!). The `number` is
-    /// the physical page number of the first page of this run. It is mutated
-    /// while we post-process the pages in this function. This function returns
-    /// a fragment consisting of multiple frames, one per output page of this
-    /// page run.
-    #[typst_macros::time(name = "page", span = self.span())]
-    pub fn layout(
-        &self,
-        engine: &mut Engine,
-        styles: StyleChain,
-        page_counter: &mut ManualPageCounter,
-        extend_to: Option<Parity>,
-    ) -> SourceResult<Vec<Page>> {
-        // When one of the lengths is infinite the page fits its content along
-        // that axis.
-        let width = self.width(styles).unwrap_or(Abs::inf());
-        let height = self.height(styles).unwrap_or(Abs::inf());
-        let mut size = Size::new(width, height);
-        if self.flipped(styles) {
-            std::mem::swap(&mut size.x, &mut size.y);
-        }
+impl Construct for PageElem {
+    fn construct(engine: &mut Engine, args: &mut Args) -> SourceResult<Content> {
+        // The page constructor is special: It doesn't create a page element.
+        // Instead, it just ensures that the passed content lives in a separate
+        // page and styles it.
+        let styles = Self::set(engine, args)?;
+        let body = args.expect::<Content>("body")?;
+        Ok(Content::sequence([
+            PagebreakElem::shared_weak().clone(),
+            // We put an effectless, invisible non-tag element on the page.
+            // This has two desirable consequences:
+            // - The page is kept even if the body is empty
+            // - The page doesn't inherit shared styles from the body
+            FlushElem::new().pack(),
+            body,
+            PagebreakElem::shared_boundary().clone(),
+        ])
+        .styled_with_map(styles))
+    }
+}
 
-        let mut min = width.min(height);
-        if !min.is_finite() {
-            min = Paper::A4.width();
-        }
+/// A manual page break.
+///
+/// Must not be used inside any containers.
+///
+/// # Example
+/// ```example
+/// The next page contains
+/// more details on compound theory.
+/// #pagebreak()
+///
+/// == Compound Theory
+/// In 1984, the first ...
+/// ```
+#[elem(title = "Page Break")]
+pub struct PagebreakElem {
+    /// If `{true}`, the page break is skipped if the current page is already
+    /// empty.
+    #[default(false)]
+    pub weak: bool,
 
-        // Determine the margins.
-        let default = Rel::<Length>::from((2.5 / 21.0) * min);
-        let margin = self.margin(styles);
-        let two_sided = margin.two_sided.unwrap_or(false);
-        let margin = margin
-            .sides
-            .map(|side| side.and_then(Smart::custom).unwrap_or(default))
-            .resolve(styles)
-            .relative_to(size);
+    /// If given, ensures that the next page will be an even/odd page, with an
+    /// empty page in between if necessary.
+    ///
+    /// ```example
+    /// #set page(height: 30pt)
+    ///
+    /// First.
+    /// #pagebreak(to: "odd")
+    /// Third.
+    /// ```
+    pub to: Option<Parity>,
 
-        // Determine the binding.
-        let binding =
-            self.binding(styles)
-                .unwrap_or_else(|| match TextElem::dir_in(styles) {
-                    Dir::LTR => Binding::Left,
-                    _ => Binding::Right,
-                });
+    /// Whether this pagebreak designates an end boundary of a page run. This is
+    /// an even weaker version of pagebreak `weak` because it not only doesn't
+    /// force an empty page, but also doesn't force its initial styles onto a
+    /// staged empty page.
+    #[internal]
+    #[parse(None)]
+    #[default(false)]
+    pub boundary: bool,
+}
 
-        // Realize columns.
-        let mut child = self.body().clone();
-        let columns = self.columns(styles);
-        if columns.get() > 1 {
-            child = ColumnsElem::new(child)
-                .with_count(columns)
-                .pack()
-                .spanned(self.span());
-        }
+impl PagebreakElem {
+    /// Get the globally shared weak pagebreak element.
+    pub fn shared_weak() -> &'static Content {
+        singleton!(Content, PagebreakElem::new().with_weak(true).pack())
+    }
 
-        let area = size - margin.sum_by_axis();
-        let mut regions = Regions::repeat(area, area.map(Abs::is_finite));
-        regions.root = true;
-
-        // Layout the child.
-        let mut frames = child.layout(engine, styles, regions)?.into_frames();
-
-        // Align the child to the pagebreak's parity.
-        // Check for page count after adding the pending frames
-        if extend_to
-            .is_some_and(|p| !p.matches(page_counter.physical().get() + frames.len()))
-        {
-            // Insert empty page after the current pages.
-            let size = area.map(Abs::is_finite).select(area, Size::zero());
-            frames.push(Frame::hard(size));
-        }
-
-        let fill = self.fill(styles);
-        let foreground = Cow::Borrowed(self.foreground(styles));
-        let background = Cow::Borrowed(self.background(styles));
-        let header_ascent = self.header_ascent(styles);
-        let footer_descent = self.footer_descent(styles);
-        let numbering = self.numbering(styles);
-        let number_align = self.number_align(styles);
-        let mut header = Cow::Borrowed(self.header(styles));
-        let mut footer = Cow::Borrowed(self.footer(styles));
-
-        // Construct the numbering (for header or footer).
-        let numbering_marginal = Cow::Owned(numbering.as_ref().map(|numbering| {
-            let both = match numbering {
-                Numbering::Pattern(pattern) => pattern.pieces() >= 2,
-                Numbering::Func(_) => true,
-            };
-
-            let mut counter = CounterDisplayElem::new(
-                Counter::new(CounterKey::Page),
-                Smart::Custom(numbering.clone()),
-                both,
-            )
-            .pack()
-            .spanned(self.span());
-
-            // We interpret the Y alignment as selecting header or footer
-            // and then ignore it for aligning the actual number.
-            if let Some(x) = number_align.x() {
-                counter = counter.aligned(x.into());
-            }
-
-            counter
-        }));
-
-        if matches!(number_align.y(), Some(OuterVAlignment::Top)) {
-            header = if header.is_some() { header } else { numbering_marginal };
-        } else {
-            footer = if footer.is_some() { footer } else { numbering_marginal };
-        }
-
-        // Post-process pages.
-        let mut pages = Vec::with_capacity(frames.len());
-        for mut frame in frames {
-            // The padded width of the page's content without margins.
-            let pw = frame.width();
-
-            // If two sided, left becomes inside and right becomes outside.
-            // Thus, for left-bound pages, we want to swap on even pages and
-            // for right-bound pages, we want to swap on odd pages.
-            let mut margin = margin;
-            if two_sided && binding.swap(page_counter.physical()) {
-                std::mem::swap(&mut margin.left, &mut margin.right);
-            }
-
-            // Realize margins.
-            frame.set_size(frame.size() + margin.sum_by_axis());
-            frame.translate(Point::new(margin.left, margin.top));
-
-            // The page size with margins.
-            let size = frame.size();
-
-            // Realize overlays.
-            for marginal in [&header, &footer, &background, &foreground] {
-                let Some(content) = &**marginal else { continue };
-
-                let (pos, area, align);
-                if ptr::eq(marginal, &header) {
-                    let ascent = header_ascent.relative_to(margin.top);
-                    pos = Point::with_x(margin.left);
-                    area = Size::new(pw, margin.top - ascent);
-                    align = Alignment::BOTTOM;
-                } else if ptr::eq(marginal, &footer) {
-                    let descent = footer_descent.relative_to(margin.bottom);
-                    pos = Point::new(margin.left, size.y - margin.bottom + descent);
-                    area = Size::new(pw, margin.bottom - descent);
-                    align = Alignment::TOP;
-                } else {
-                    pos = Point::zero();
-                    area = size;
-                    align = HAlignment::Center + VAlignment::Horizon;
-                };
-
-                let pod = Regions::one(area, Axes::splat(true));
-                let sub = content
-                    .clone()
-                    .styled(AlignElem::set_alignment(align))
-                    .layout(engine, styles, pod)?
-                    .into_frame();
-
-                if ptr::eq(marginal, &header) || ptr::eq(marginal, &background) {
-                    frame.prepend_frame(pos, sub);
-                } else {
-                    frame.push_frame(pos, sub);
-                }
-            }
-
-            if let Some(fill) = fill {
-                frame.fill(fill.clone());
-            }
-
-            page_counter.visit(engine, &frame)?;
-            pages.push(Page {
-                frame,
-                numbering: numbering.clone(),
-                number: page_counter.logical(),
-            });
-
-            page_counter.step();
-        }
-
-        Ok(pages)
+    /// Get the globally shared boundary pagebreak element.
+    pub fn shared_boundary() -> &'static Content {
+        singleton!(
+            Content,
+            PagebreakElem::new().with_weak(true).with_boundary(true).pack()
+        )
     }
 }
 
@@ -520,6 +438,15 @@ impl Packed<PageElem> {
 pub struct Page {
     /// The frame that defines the page.
     pub frame: Frame,
+    /// How the page is filled.
+    ///
+    /// - When `None`, the background is transparent.
+    /// - When `Auto`, the background is transparent for PDF and white
+    ///   for raster and SVG targets.
+    ///
+    /// Exporters should access the resolved value of this property through
+    /// `fill_or_transparent()` or `fill_or_white()`.
+    pub fill: Smart<Option<Paint>>,
     /// The page's numbering.
     pub numbering: Option<Numbering>,
     /// The logical page number (controlled by `counter(page)` and may thus not
@@ -527,6 +454,22 @@ pub struct Page {
     pub number: usize,
 }
 
+impl Page {
+    /// Get the configured background or `None` if it is `Auto`.
+    ///
+    /// This is used in PDF export.
+    pub fn fill_or_transparent(&self) -> Option<Paint> {
+        self.fill.clone().unwrap_or(None)
+    }
+
+    /// Get the configured background or white if it is `Auto`.
+    ///
+    /// This is used in raster and SVG export.
+    pub fn fill_or_white(&self) -> Option<Paint> {
+        self.fill.clone().unwrap_or_else(|| Some(Color::WHITE.into()))
+    }
+}
+
 /// Specification of the page's margins.
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
 pub struct Margin {
@@ -648,7 +591,7 @@ pub enum Binding {
 
 impl Binding {
     /// Whether to swap left and right margin for the page with this number.
-    fn swap(self, number: NonZeroUsize) -> bool {
+    pub fn swap(self, number: NonZeroUsize) -> bool {
         match self {
             // Left-bound must swap on even pages
             // (because it is correct on the first page).
@@ -710,37 +653,41 @@ cast! {
     v: Func => Self::Func(v),
 }
 
-/// A manual page break.
-///
-/// Must not be used inside any containers.
-///
-/// # Example
-/// ```example
-/// The next page contains
-/// more details on compound theory.
-/// #pagebreak()
-///
-/// == Compound Theory
-/// In 1984, the first ...
-/// ```
-#[elem(title = "Page Break")]
-pub struct PagebreakElem {
-    /// If `{true}`, the page break is skipped if the current page is already
-    /// empty.
-    #[default(false)]
-    pub weak: bool,
+/// A list of page ranges to be exported.
+#[derive(Debug, Clone)]
+pub struct PageRanges(Vec<PageRange>);
 
-    /// If given, ensures that the next page will be an even/odd page, with an
-    /// empty page in between if necessary.
-    ///
-    /// ```example
-    /// #set page(height: 30pt)
-    ///
-    /// First.
-    /// #pagebreak(to: "odd")
-    /// Third.
-    /// ```
-    pub to: Option<Parity>,
+/// A range of pages to export.
+///
+/// The range is one-indexed. For example, `1..=3` indicates the first, second
+/// and third pages should be exported.
+pub type PageRange = RangeInclusive<Option<NonZeroUsize>>;
+
+impl PageRanges {
+    /// Create new page ranges.
+    pub fn new(ranges: Vec<PageRange>) -> Self {
+        Self(ranges)
+    }
+
+    /// Check if a page, given its number, should be included when exporting the
+    /// document while restricting the exported pages to these page ranges.
+    /// This is the one-indexed version of 'includes_page_index'.
+    pub fn includes_page(&self, page: NonZeroUsize) -> bool {
+        self.includes_page_index(page.get() - 1)
+    }
+
+    /// Check if a page, given its index, should be included when exporting the
+    /// document while restricting the exported pages to these page ranges.
+    /// This is the zero-indexed version of 'includes_page'.
+    pub fn includes_page_index(&self, page: usize) -> bool {
+        let page = NonZeroUsize::try_from(page + 1).unwrap();
+        self.0.iter().any(|range| match (range.start(), range.end()) {
+            (Some(start), Some(end)) => (start..=end).contains(&&page),
+            (Some(start), None) => (start..).contains(&&page),
+            (None, Some(end)) => (..=end).contains(&&page),
+            (None, None) => true,
+        })
+    }
 }
 
 /// Whether something should be even or odd.
@@ -754,7 +701,7 @@ pub enum Parity {
 
 impl Parity {
     /// Whether the given number matches the parity.
-    fn matches(self, number: usize) -> bool {
+    pub fn matches(self, number: usize) -> bool {
         match self {
             Self::Even => number % 2 == 0,
             Self::Odd => number % 2 == 1,
