@@ -1,14 +1,22 @@
-use crate::diag::SourceResult;
+use std::ops::Div;
+
+use once_cell::unsync::Lazy;
+
+use crate::diag::{bail, SourceResult};
 use crate::engine::Engine;
-use crate::foundations::{elem, Content, Packed, Resolve, StyleChain};
-use crate::layout::{
-    Abs, Alignment, Angle, Axes, FixedAlignment, Frame, HAlignment, LayoutMultiple,
-    LayoutSingle, Length, Point, Ratio, Regions, Rel, Size, VAlignment,
+use crate::foundations::{
+    cast, elem, Content, NativeElement, Packed, Resolve, Show, Smart, StyleChain,
 };
+use crate::introspection::Locator;
+use crate::layout::{
+    layout_frame, Abs, Alignment, Angle, Axes, BlockElem, FixedAlignment, Frame,
+    HAlignment, Length, Point, Ratio, Region, Rel, Size, VAlignment,
+};
+use crate::utils::Numeric;
 
 /// Moves content without affecting layout.
 ///
-/// The `move` function allows you to move content while th layout still 'sees'
+/// The `move` function allows you to move content while the layout still 'sees'
 /// it at the original positions. Containers will still be sized as if the
 /// content was not moved.
 ///
@@ -24,7 +32,7 @@ use crate::layout::{
 ///   )
 /// ))
 /// ```
-#[elem(LayoutSingle)]
+#[elem(Show)]
 pub struct MoveElem {
     /// The horizontal displacement of the content.
     pub dx: Rel<Length>,
@@ -37,23 +45,30 @@ pub struct MoveElem {
     pub body: Content,
 }
 
-impl LayoutSingle for Packed<MoveElem> {
-    #[typst_macros::time(name = "move", span = self.span())]
-    fn layout(
-        &self,
-        engine: &mut Engine,
-        styles: StyleChain,
-        regions: Regions,
-    ) -> SourceResult<Frame> {
-        let pod = Regions::one(regions.base(), Axes::splat(false));
-        let mut frame = self.body().layout(engine, styles, pod)?.into_frame();
-        let delta = Axes::new(self.dx(styles), self.dy(styles)).resolve(styles);
-        let delta = delta.zip_map(regions.base(), Rel::relative_to);
-        frame.translate(delta.to_point());
-        Ok(frame)
+impl Show for Packed<MoveElem> {
+    fn show(&self, _: &mut Engine, _: StyleChain) -> SourceResult<Content> {
+        Ok(BlockElem::single_layouter(self.clone(), layout_move)
+            .pack()
+            .spanned(self.span()))
     }
 }
 
+/// Layout the moved content.
+#[typst_macros::time(span = elem.span())]
+fn layout_move(
+    elem: &Packed<MoveElem>,
+    engine: &mut Engine,
+    locator: Locator,
+    styles: StyleChain,
+    region: Region,
+) -> SourceResult<Frame> {
+    let mut frame = layout_frame(engine, &elem.body, locator, styles, region)?;
+    let delta = Axes::new(elem.dx(styles), elem.dy(styles)).resolve(styles);
+    let delta = delta.zip_map(region.size, Rel::relative_to);
+    frame.translate(delta.to_point());
+    Ok(frame)
+}
+
 /// Rotates content without affecting layout.
 ///
 /// Rotates an element by a given angle. The layout will act as if the element
@@ -68,7 +83,7 @@ impl LayoutSingle for Packed<MoveElem> {
 ///     .map(i => rotate(24deg * i)[X]),
 /// )
 /// ```
-#[elem(LayoutSingle)]
+#[elem(Show)]
 pub struct RotateElem {
     /// The amount of rotation.
     ///
@@ -115,38 +130,46 @@ pub struct RotateElem {
     pub body: Content,
 }
 
-impl LayoutSingle for Packed<RotateElem> {
-    #[typst_macros::time(name = "rotate", span = self.span())]
-    fn layout(
-        &self,
-        engine: &mut Engine,
-        styles: StyleChain,
-        regions: Regions,
-    ) -> SourceResult<Frame> {
-        let angle = self.angle(styles);
-        let align = self.origin(styles).resolve(styles);
-
-        // Compute the new region's approximate size.
-        let size = regions
-            .base()
-            .to_point()
-            .transform_inf(Transform::rotate(angle))
-            .map(Abs::abs)
-            .to_size();
-
-        measure_and_layout(
-            engine,
-            regions.base(),
-            size,
-            styles,
-            self.body(),
-            Transform::rotate(angle),
-            align,
-            self.reflow(styles),
-        )
+impl Show for Packed<RotateElem> {
+    fn show(&self, _: &mut Engine, _: StyleChain) -> SourceResult<Content> {
+        Ok(BlockElem::single_layouter(self.clone(), layout_rotate)
+            .pack()
+            .spanned(self.span()))
     }
 }
 
+/// Layout the rotated content.
+#[typst_macros::time(span = elem.span())]
+fn layout_rotate(
+    elem: &Packed<RotateElem>,
+    engine: &mut Engine,
+    locator: Locator,
+    styles: StyleChain,
+    region: Region,
+) -> SourceResult<Frame> {
+    let angle = elem.angle(styles);
+    let align = elem.origin(styles).resolve(styles);
+
+    // Compute the new region's approximate size.
+    let size = if region.size.is_finite() {
+        compute_bounding_box(region.size, Transform::rotate(-angle)).1
+    } else {
+        Size::splat(Abs::inf())
+    };
+
+    measure_and_layout(
+        engine,
+        locator,
+        region,
+        size,
+        styles,
+        elem.body(),
+        Transform::rotate(angle),
+        align,
+        elem.reflow(styles),
+    )
+}
+
 /// Scales content without affecting layout.
 ///
 /// Lets you mirror content by specifying a negative scale on a single axis.
@@ -157,8 +180,16 @@ impl LayoutSingle for Packed<RotateElem> {
 /// #scale(x: -100%)[This is mirrored.]
 /// #scale(x: -100%, reflow: true)[This is mirrored.]
 /// ```
-#[elem(LayoutSingle)]
+#[elem(Show)]
 pub struct ScaleElem {
+    /// The scaling factor for both axes, as a positional argument. This is just
+    /// an optional shorthand notation for setting `x` and `y` to the same
+    /// value.
+    #[external]
+    #[positional]
+    #[default(Smart::Custom(ScaleAmount::Ratio(Ratio::one())))]
+    pub factor: Smart<ScaleAmount>,
+
     /// The horizontal scaling factor.
     ///
     /// The body will be mirrored horizontally if the parameter is negative.
@@ -166,15 +197,15 @@ pub struct ScaleElem {
         let all = args.find()?;
         args.named("x")?.or(all)
     )]
-    #[default(Ratio::one())]
-    pub x: Ratio,
+    #[default(Smart::Custom(ScaleAmount::Ratio(Ratio::one())))]
+    pub x: Smart<ScaleAmount>,
 
     /// The vertical scaling factor.
     ///
     /// The body will be mirrored vertically if the parameter is negative.
     #[parse(args.named("y")?.or(all))]
-    #[default(Ratio::one())]
-    pub y: Ratio,
+    #[default(Smart::Custom(ScaleAmount::Ratio(Ratio::one())))]
+    pub y: Smart<ScaleAmount>,
 
     /// The origin of the transformation.
     ///
@@ -203,37 +234,217 @@ pub struct ScaleElem {
     pub body: Content,
 }
 
-impl LayoutSingle for Packed<ScaleElem> {
-    #[typst_macros::time(name = "scale", span = self.span())]
-    fn layout(
+impl Show for Packed<ScaleElem> {
+    fn show(&self, _: &mut Engine, _: StyleChain) -> SourceResult<Content> {
+        Ok(BlockElem::single_layouter(self.clone(), layout_scale)
+            .pack()
+            .spanned(self.span()))
+    }
+}
+
+/// Layout the scaled content.
+#[typst_macros::time(span = elem.span())]
+fn layout_scale(
+    elem: &Packed<ScaleElem>,
+    engine: &mut Engine,
+    locator: Locator,
+    styles: StyleChain,
+    region: Region,
+) -> SourceResult<Frame> {
+    // Compute the new region's approximate size.
+    let scale = elem.resolve_scale(engine, locator.relayout(), region.size, styles)?;
+    let size = region
+        .size
+        .zip_map(scale, |r, s| if r.is_finite() { Ratio::new(1.0 / s).of(r) } else { r })
+        .map(Abs::abs);
+
+    measure_and_layout(
+        engine,
+        locator,
+        region,
+        size,
+        styles,
+        elem.body(),
+        Transform::scale(scale.x, scale.y),
+        elem.origin(styles).resolve(styles),
+        elem.reflow(styles),
+    )
+}
+
+#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
+enum ScaleAmount {
+    Ratio(Ratio),
+    Length(Length),
+}
+
+impl Packed<ScaleElem> {
+    /// Resolves scale parameters, preserving aspect ratio if one of the scales is set to `auto`.
+    fn resolve_scale(
         &self,
         engine: &mut Engine,
+        locator: Locator,
+        container: Size,
         styles: StyleChain,
-        regions: Regions,
-    ) -> SourceResult<Frame> {
-        let sx = self.x(styles);
-        let sy = self.y(styles);
-        let align = self.origin(styles).resolve(styles);
+    ) -> SourceResult<Axes<Ratio>> {
+        fn resolve_axis(
+            axis: Smart<ScaleAmount>,
+            body: impl Fn() -> SourceResult<Abs>,
+            styles: StyleChain,
+        ) -> SourceResult<Smart<Ratio>> {
+            Ok(match axis {
+                Smart::Auto => Smart::Auto,
+                Smart::Custom(amt) => Smart::Custom(match amt {
+                    ScaleAmount::Ratio(ratio) => ratio,
+                    ScaleAmount::Length(length) => {
+                        let length = length.resolve(styles);
+                        Ratio::new(length.div(body()?))
+                    }
+                }),
+            })
+        }
 
-        // Compute the new region's approximate size.
-        let size = regions
-            .base()
-            .zip_map(Axes::new(sx, sy), |r, s| s.of(r))
-            .map(Abs::abs);
+        let size = Lazy::new(|| {
+            let pod = Region::new(container, Axes::splat(false));
+            let frame = layout_frame(engine, &self.body, locator, styles, pod)?;
+            SourceResult::Ok(frame.size())
+        });
 
-        measure_and_layout(
-            engine,
-            regions.base(),
-            size,
+        let x = resolve_axis(
+            self.x(styles),
+            || size.as_ref().map(|size| size.x).map_err(Clone::clone),
             styles,
-            self.body(),
-            Transform::scale(sx, sy),
-            align,
-            self.reflow(styles),
-        )
+        )?;
+
+        let y = resolve_axis(
+            self.y(styles),
+            || size.as_ref().map(|size| size.y).map_err(Clone::clone),
+            styles,
+        )?;
+
+        match (x, y) {
+            (Smart::Auto, Smart::Auto) => {
+                bail!(self.span(), "x and y cannot both be auto")
+            }
+            (Smart::Custom(x), Smart::Custom(y)) => Ok(Axes::new(x, y)),
+            (Smart::Auto, Smart::Custom(v)) | (Smart::Custom(v), Smart::Auto) => {
+                Ok(Axes::splat(v))
+            }
+        }
     }
 }
 
+cast! {
+    ScaleAmount,
+    self => match self {
+        ScaleAmount::Ratio(ratio) => ratio.into_value(),
+        ScaleAmount::Length(length) => length.into_value(),
+    },
+    ratio: Ratio => ScaleAmount::Ratio(ratio),
+    length: Length => ScaleAmount::Length(length),
+}
+
+/// Skews content.
+///
+/// Skews an element in horizontal and/or vertical direction. The layout will
+/// act as if the element was not skewed unless you specify `{reflow: true}`.
+///
+/// # Example
+/// ```example
+/// #skew(ax: -12deg)[
+///   This is some fake italic text.
+/// ]
+/// ```
+#[elem(Show)]
+pub struct SkewElem {
+    /// The horizontal skewing angle.
+    ///
+    /// ```example
+    /// #skew(ax: 30deg)[Skewed]
+    /// ```
+    ///
+    #[default(Angle::zero())]
+    pub ax: Angle,
+
+    /// The vertical skewing angle.
+    ///
+    /// ```example
+    /// #skew(ay: 30deg)[Skewed]
+    /// ```
+    ///
+    #[default(Angle::zero())]
+    pub ay: Angle,
+
+    /// The origin of the skew transformation.
+    ///
+    /// The origin will stay fixed during the operation.
+    ///
+    /// ```example
+    /// X #box(skew(ax: -30deg, origin: center + horizon)[X]) X \
+    /// X #box(skew(ax: -30deg, origin: bottom + left)[X]) X \
+    /// X #box(skew(ax: -30deg, origin: top + right)[X]) X
+    /// ```
+    #[fold]
+    #[default(HAlignment::Center + VAlignment::Horizon)]
+    pub origin: Alignment,
+
+    /// Whether the skew transformation impacts the layout.
+    ///
+    /// If set to `{false}`, the skewed content will retain the bounding box of
+    /// the original content. If set to `{true}`, the bounding box will take the
+    /// transformation of the content into account and adjust the layout accordingly.
+    ///
+    /// ```example
+    /// Hello #skew(ay: 30deg, reflow: true, "World")!
+    /// ```
+    #[default(false)]
+    pub reflow: bool,
+
+    /// The content to skew.
+    #[required]
+    pub body: Content,
+}
+
+impl Show for Packed<SkewElem> {
+    fn show(&self, _: &mut Engine, _: StyleChain) -> SourceResult<Content> {
+        Ok(BlockElem::single_layouter(self.clone(), layout_skew)
+            .pack()
+            .spanned(self.span()))
+    }
+}
+
+/// Layout the skewed content.
+#[typst_macros::time(span = elem.span())]
+fn layout_skew(
+    elem: &Packed<SkewElem>,
+    engine: &mut Engine,
+    locator: Locator,
+    styles: StyleChain,
+    region: Region,
+) -> SourceResult<Frame> {
+    let ax = elem.ax(styles);
+    let ay = elem.ay(styles);
+    let align = elem.origin(styles).resolve(styles);
+
+    // Compute the new region's approximate size.
+    let size = if region.size.is_finite() {
+        compute_bounding_box(region.size, Transform::skew(ax, ay)).1
+    } else {
+        Size::splat(Abs::inf())
+    };
+
+    measure_and_layout(
+        engine,
+        locator,
+        region,
+        size,
+        styles,
+        elem.body(),
+        Transform::skew(ax, ay),
+        align,
+        elem.reflow(styles),
+    )
+}
+
 /// A scale-skew-translate transformation.
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
 pub struct Transform {
@@ -281,6 +492,15 @@ impl Transform {
         }
     }
 
+    /// A skew transform.
+    pub fn skew(ax: Angle, ay: Angle) -> Self {
+        Self {
+            kx: Ratio::new(ax.tan()),
+            ky: Ratio::new(ay.tan()),
+            ..Self::identity()
+        }
+    }
+
     /// Whether this is the identity transformation.
     pub fn is_identity(self) -> bool {
         self == Self::identity()
@@ -363,7 +583,8 @@ impl Default for Transform {
 #[allow(clippy::too_many_arguments)]
 fn measure_and_layout(
     engine: &mut Engine,
-    base_size: Size,
+    locator: Locator,
+    region: Region,
     size: Size,
     styles: StyleChain,
     body: &Content,
@@ -371,57 +592,57 @@ fn measure_and_layout(
     align: Axes<FixedAlignment>,
     reflow: bool,
 ) -> SourceResult<Frame> {
-    if !reflow {
-        // Layout the body.
-        let pod = Regions::one(base_size, Axes::splat(false));
-        let mut frame = body.layout(engine, styles, pod)?.into_frame();
+    if reflow {
+        // Measure the size of the body.
+        let pod = Region::new(size, Axes::splat(false));
+        let frame = layout_frame(engine, body, locator.relayout(), styles, pod)?;
+
+        // Actually perform the layout.
+        let pod = Region::new(frame.size(), Axes::splat(true));
+        let mut frame = layout_frame(engine, body, locator, styles, pod)?;
         let Axes { x, y } = align.zip_map(frame.size(), FixedAlignment::position);
 
-        // Apply the transform.
+        // Compute the transform.
         let ts = Transform::translate(x, y)
             .pre_concat(transform)
             .pre_concat(Transform::translate(-x, -y));
+
+        // Compute the bounding box and offset and wrap in a new frame.
+        let (offset, size) = compute_bounding_box(frame.size(), ts);
         frame.transform(ts);
+        frame.translate(offset);
+        frame.set_size(size);
+        Ok(frame)
+    } else {
+        // Layout the body.
+        let mut frame = layout_frame(engine, body, locator, styles, region)?;
+        let Axes { x, y } = align.zip_map(frame.size(), FixedAlignment::position);
 
-        return Ok(frame);
+        // Compute the transform.
+        let ts = Transform::translate(x, y)
+            .pre_concat(transform)
+            .pre_concat(Transform::translate(-x, -y));
+
+        // Apply the transform.
+        frame.transform(ts);
+        Ok(frame)
     }
-
-    // Measure the size of the body.
-    let pod = Regions::one(size, Axes::splat(false));
-    let frame = body.measure(engine, styles, pod)?.into_frame();
-
-    // Actually perform the layout.
-    let pod = Regions::one(frame.size(), Axes::splat(true));
-    let mut frame = body.layout(engine, styles, pod)?.into_frame();
-    let Axes { x, y } = align.zip_map(frame.size(), FixedAlignment::position);
-
-    // Apply the transform.
-    let ts = Transform::translate(x, y)
-        .pre_concat(transform)
-        .pre_concat(Transform::translate(-x, -y));
-
-    // Compute the bounding box and offset and wrap in a new frame.
-    let (offset, size) = compute_bounding_box(&frame, ts);
-    frame.transform(ts);
-    frame.translate(offset);
-    frame.set_size(size);
-    Ok(frame)
 }
 
-/// Computes the bounding box and offset of a transformed frame.
-fn compute_bounding_box(frame: &Frame, ts: Transform) -> (Point, Size) {
+/// Computes the bounding box and offset of a transformed area.
+fn compute_bounding_box(size: Size, ts: Transform) -> (Point, Size) {
     let top_left = Point::zero().transform_inf(ts);
-    let top_right = Point::new(frame.width(), Abs::zero()).transform_inf(ts);
-    let bottom_left = Point::new(Abs::zero(), frame.height()).transform_inf(ts);
-    let bottom_right = Point::new(frame.width(), frame.height()).transform_inf(ts);
+    let top_right = Point::with_x(size.x).transform_inf(ts);
+    let bottom_left = Point::with_y(size.y).transform_inf(ts);
+    let bottom_right = size.to_point().transform_inf(ts);
 
-    // We first compute the new bounding box of the rotated frame.
+    // We first compute the new bounding box of the rotated area.
     let min_x = top_left.x.min(top_right.x).min(bottom_left.x).min(bottom_right.x);
     let min_y = top_left.y.min(top_right.y).min(bottom_left.y).min(bottom_right.y);
     let max_x = top_left.x.max(top_right.x).max(bottom_left.x).max(bottom_right.x);
     let max_y = top_left.y.max(top_right.y).max(bottom_left.y).max(bottom_right.y);
 
-    // Then we compute the new size of the frame.
+    // Then we compute the new size of the area.
     let width = max_x - min_x;
     let height = max_y - min_y;
 
