@@ -1,6 +1,6 @@
 use std::f64::consts::SQRT_2;
 
-use ecow::EcoString;
+use ecow::{eco_vec, EcoString};
 use rustybuzz::Feature;
 use ttf_parser::gsub::{AlternateSubstitution, SingleSubstitution, SubstitutionSubtable};
 use ttf_parser::math::MathValue;
@@ -11,17 +11,20 @@ use unicode_segmentation::UnicodeSegmentation;
 
 use crate::diag::SourceResult;
 use crate::engine::Engine;
-use crate::foundations::{Content, Packed, StyleChain};
-use crate::layout::{Abs, Axes, BoxElem, Em, Frame, LayoutMultiple, Regions, Size};
+use crate::foundations::{Content, Packed, StyleChain, StyleVec};
+use crate::introspection::{SplitLocator, TagElem};
+use crate::layout::{
+    layout_frame, Abs, Axes, BoxElem, Em, Frame, HElem, PlaceElem, Region, Size, Spacing,
+};
 use crate::math::{
     scaled_font_size, styled_char, EquationElem, FrameFragment, GlyphFragment,
     LayoutMath, MathFragment, MathRun, MathSize, THICK,
 };
-use crate::model::ParElem;
+use crate::realize::{realize, Arenas, RealizationKind};
 use crate::syntax::{is_newline, Span};
 use crate::text::{
-    features, BottomEdge, BottomEdgeMetric, Font, TextElem, TextSize, TopEdge,
-    TopEdgeMetric,
+    features, BottomEdge, BottomEdgeMetric, Font, LinebreakElem, SpaceElem, TextElem,
+    TextSize, TopEdge, TopEdgeMetric,
 };
 
 macro_rules! scaled {
@@ -45,10 +48,11 @@ macro_rules! percent {
 }
 
 /// The context for math layout.
-pub struct MathContext<'a, 'b, 'v> {
+pub struct MathContext<'a, 'v, 'e> {
     // External.
-    pub engine: &'v mut Engine<'b>,
-    pub regions: Regions<'static>,
+    pub engine: &'v mut Engine<'e>,
+    pub locator: &'v mut SplitLocator<'a>,
+    pub region: Region,
     // Font-related.
     pub font: &'a Font,
     pub ttf: &'a ttf_parser::Face<'a>,
@@ -61,11 +65,13 @@ pub struct MathContext<'a, 'b, 'v> {
     pub fragments: Vec<MathFragment>,
 }
 
-impl<'a, 'b, 'v> MathContext<'a, 'b, 'v> {
+impl<'a, 'v, 'e> MathContext<'a, 'v, 'e> {
+    /// Create a new math context.
     pub fn new(
-        engine: &'v mut Engine<'b>,
+        engine: &'v mut Engine<'e>,
+        locator: &'v mut SplitLocator<'a>,
         styles: StyleChain<'a>,
-        regions: Regions,
+        base: Size,
         font: &'a Font,
     ) -> Self {
         let math_table = font.ttf().tables().math.unwrap();
@@ -102,7 +108,8 @@ impl<'a, 'b, 'v> MathContext<'a, 'b, 'v> {
 
         Self {
             engine,
-            regions: Regions::one(regions.base(), Axes::splat(false)),
+            locator,
+            region: Region::new(base, Axes::splat(false)),
             font,
             ttf: font.ttf(),
             table: math_table,
@@ -114,18 +121,29 @@ impl<'a, 'b, 'v> MathContext<'a, 'b, 'v> {
         }
     }
 
+    /// Push a fragment.
     pub fn push(&mut self, fragment: impl Into<MathFragment>) {
         self.fragments.push(fragment.into());
     }
 
-    pub fn extend(&mut self, fragments: Vec<MathFragment>) {
+    /// Push multiple fragments.
+    pub fn extend(&mut self, fragments: impl IntoIterator<Item = MathFragment>) {
         self.fragments.extend(fragments);
     }
 
+    /// Layout the given element and return the result as a [`MathRun`].
+    pub fn layout_into_run(
+        &mut self,
+        elem: &Content,
+        styles: StyleChain,
+    ) -> SourceResult<MathRun> {
+        Ok(MathRun::new(self.layout_into_fragments(elem, styles)?))
+    }
+
     /// Layout the given element and return the resulting [`MathFragment`]s.
     pub fn layout_into_fragments(
         &mut self,
-        elem: &dyn LayoutMath,
+        elem: &Content,
         styles: StyleChain,
     ) -> SourceResult<Vec<MathFragment>> {
         // The element's layout_math() changes the fragments held in this
@@ -133,24 +151,15 @@ impl<'a, 'b, 'v> MathContext<'a, 'b, 'v> {
         // them, so we restore the MathContext's fragments after obtaining the
         // layout result.
         let prev = std::mem::take(&mut self.fragments);
-        elem.layout_math(self, styles)?;
+        self.layout(elem, styles)?;
         Ok(std::mem::replace(&mut self.fragments, prev))
     }
 
-    /// Layout the given element and return the result as a [`MathRun`].
-    pub fn layout_into_run(
-        &mut self,
-        elem: &dyn LayoutMath,
-        styles: StyleChain,
-    ) -> SourceResult<MathRun> {
-        Ok(MathRun::new(self.layout_into_fragments(elem, styles)?))
-    }
-
     /// Layout the given element and return the result as a
     /// unified [`MathFragment`].
     pub fn layout_into_fragment(
         &mut self,
-        elem: &dyn LayoutMath,
+        elem: &Content,
         styles: StyleChain,
     ) -> SourceResult<MathFragment> {
         Ok(self.layout_into_run(elem, styles)?.into_fragment(self, styles))
@@ -159,38 +168,122 @@ impl<'a, 'b, 'v> MathContext<'a, 'b, 'v> {
     /// Layout the given element and return the result as a [`Frame`].
     pub fn layout_into_frame(
         &mut self,
-        elem: &dyn LayoutMath,
+        elem: &Content,
         styles: StyleChain,
     ) -> SourceResult<Frame> {
         Ok(self.layout_into_fragment(elem, styles)?.into_frame())
     }
+}
 
-    /// Layout the given [`BoxElem`] into a [`Frame`].
-    pub fn layout_box(
+impl MathContext<'_, '_, '_> {
+    /// Layout arbitrary content.
+    fn layout(&mut self, content: &Content, styles: StyleChain) -> SourceResult<()> {
+        let arenas = Arenas::default();
+        let pairs = realize(
+            RealizationKind::Math,
+            self.engine,
+            self.locator,
+            &arenas,
+            content,
+            styles,
+        )?;
+
+        let outer = styles;
+        for (elem, styles) in pairs {
+            // Hack because the font is fixed in math.
+            if styles != outer && TextElem::font_in(styles) != TextElem::font_in(outer) {
+                let frame = self.layout_external(elem, styles)?;
+                self.push(FrameFragment::new(self, styles, frame).with_spaced(true));
+                continue;
+            }
+
+            self.layout_realized(elem, styles)?;
+        }
+
+        Ok(())
+    }
+
+    /// Layout an element resulting from realization.
+    fn layout_realized(
+        &mut self,
+        elem: &Content,
+        styles: StyleChain,
+    ) -> SourceResult<()> {
+        if let Some(elem) = elem.to_packed::<TagElem>() {
+            self.push(MathFragment::Tag(elem.tag.clone()));
+        } else if elem.is::<SpaceElem>() {
+            let font_size = scaled_font_size(self, styles);
+            self.push(MathFragment::Space(self.space_width.at(font_size)));
+        } else if elem.is::<LinebreakElem>() {
+            self.push(MathFragment::Linebreak);
+        } else if let Some(elem) = elem.to_packed::<HElem>() {
+            if let Spacing::Rel(rel) = elem.amount() {
+                if rel.rel.is_zero() {
+                    self.push(MathFragment::Spacing(
+                        rel.abs.at(scaled_font_size(self, styles)),
+                        elem.weak(styles),
+                    ));
+                }
+            }
+        } else if let Some(elem) = elem.to_packed::<TextElem>() {
+            let fragment = self.layout_text(elem, styles)?;
+            self.push(fragment);
+        } else if let Some(boxed) = elem.to_packed::<BoxElem>() {
+            let frame = self.layout_box(boxed, styles)?;
+            self.push(FrameFragment::new(self, styles, frame).with_spaced(true));
+        } else if let Some(elem) = elem.with::<dyn LayoutMath>() {
+            elem.layout_math(self, styles)?;
+        } else {
+            let mut frame = self.layout_external(elem, styles)?;
+            if !frame.has_baseline() {
+                let axis = scaled!(self, styles, axis_height);
+                frame.set_baseline(frame.height() / 2.0 + axis);
+            }
+            self.push(
+                FrameFragment::new(self, styles, frame)
+                    .with_spaced(true)
+                    .with_ignorant(elem.is::<PlaceElem>()),
+            );
+        }
+
+        Ok(())
+    }
+
+    /// Layout a box into a frame.
+    fn layout_box(
         &mut self,
         boxed: &Packed<BoxElem>,
         styles: StyleChain,
     ) -> SourceResult<Frame> {
         let local =
             TextElem::set_size(TextSize(scaled_font_size(self, styles).into())).wrap();
-        boxed.layout(self.engine, styles.chain(&local), self.regions)
+        boxed.layout(
+            self.engine,
+            self.locator.next(&boxed.span()),
+            styles.chain(&local),
+            self.region.size,
+        )
     }
 
-    /// Layout the given [`Content`] into a [`Frame`].
-    pub fn layout_content(
+    /// Layout into a frame with normal layout.
+    fn layout_external(
         &mut self,
         content: &Content,
         styles: StyleChain,
     ) -> SourceResult<Frame> {
         let local =
             TextElem::set_size(TextSize(scaled_font_size(self, styles).into())).wrap();
-        Ok(content
-            .layout(self.engine, styles.chain(&local), self.regions)?
-            .into_frame())
+        layout_frame(
+            self.engine,
+            content,
+            self.locator.next(&content.span()),
+            styles.chain(&local),
+            self.region,
+        )
     }
 
     /// Layout the given [`TextElem`] into a [`MathFragment`].
-    pub fn layout_text(
+    fn layout_text(
         &mut self,
         elem: &Packed<TextElem>,
         styles: StyleChain,
@@ -285,12 +378,17 @@ impl<'a, 'b, 'v> MathContext<'a, 'b, 'v> {
         // it will overflow. So emulate an `hbox` instead and allow the paragraph
         // to extend as far as needed.
         let spaced = text.graphemes(true).nth(1).is_some();
-        let text = TextElem::packed(text).spanned(span);
-        let par = ParElem::new(vec![text]);
-        let frame = Packed::new(par)
-            .spanned(span)
-            .layout(self.engine, styles, false, Size::splat(Abs::inf()), false)?
-            .into_frame();
+        let elem = TextElem::packed(text).spanned(span);
+        let frame = crate::layout::layout_inline(
+            self.engine,
+            &StyleVec::wrap(eco_vec![elem]),
+            self.locator.next(&span),
+            styles,
+            false,
+            Size::splat(Abs::inf()),
+            false,
+        )?
+        .into_frame();
 
         Ok(FrameFragment::new(self, styles, frame)
             .with_class(MathClass::Alphabetic)
@@ -299,6 +397,7 @@ impl<'a, 'b, 'v> MathContext<'a, 'b, 'v> {
     }
 }
 
+/// Converts some unit to an absolute length with the current font & font size.
 pub(super) trait Scaled {
     fn scaled(self, ctx: &MathContext, font_size: Abs) -> Abs;
 }
