@@ -10,13 +10,15 @@ use crate::foundations::{
     NativeElement, Packed, Show, ShowSet, Smart, StyleChain, Styles,
 };
 use crate::introspection::{Counter, CounterKey, Locatable};
-use crate::layout::{BoxElem, Em, Fr, HElem, HideElem, Length, Rel, RepeatElem, Spacing};
+use crate::layout::{
+    BoxElem, Dir, Em, Fr, HElem, HideElem, Length, Rel, RepeatElem, Spacing,
+};
 use crate::model::{
     Destination, HeadingElem, NumberingPattern, ParElem, ParbreakElem, Refable,
 };
 use crate::syntax::Span;
-use crate::text::{Lang, LinebreakElem, LocalName, Region, SpaceElem, TextElem};
-use crate::util::{option_eq, NonZeroExt};
+use crate::text::{LinebreakElem, LocalName, SpaceElem, TextElem};
+use crate::utils::NonZeroExt;
 
 /// A table of contents, figures, or other elements.
 ///
@@ -73,8 +75,7 @@ pub struct OutlineElem {
     /// The outline's heading will not be numbered by default, but you can
     /// force it to be with a show-set rule:
     /// `{show outline: set heading(numbering: "1.")}`
-    #[default(Some(Smart::Auto))]
-    pub title: Option<Smart<Content>>,
+    pub title: Smart<Option<Content>>,
 
     /// The type of element to include in the outline.
     ///
@@ -171,7 +172,7 @@ pub struct OutlineElem {
     pub indent: Option<Smart<OutlineIndent>>,
 
     /// Content to fill the space between the title and the page number. Can be
-    /// set to `none` to disable filling.
+    /// set to `{none}` to disable filling.
     ///
     /// ```example
     /// #outline(fill: line(length: 100%))
@@ -191,13 +192,11 @@ impl OutlineElem {
 impl Show for Packed<OutlineElem> {
     #[typst_macros::time(name = "outline", span = self.span())]
     fn show(&self, engine: &mut Engine, styles: StyleChain) -> SourceResult<Content> {
-        let mut seq = vec![ParbreakElem::new().pack()];
+        let mut seq = vec![ParbreakElem::shared().clone()];
         // Build the outline title.
-        if let Some(title) = self.title(styles) {
-            let title = title.unwrap_or_else(|| {
-                TextElem::packed(Self::local_name_in(styles)).spanned(self.span())
-            });
-
+        if let Some(title) = self.title(styles).unwrap_or_else(|| {
+            Some(TextElem::packed(Self::local_name_in(styles)).spanned(self.span()))
+        }) {
             seq.push(
                 HeadingElem::new(title)
                     .with_depth(NonZeroUsize::ONE)
@@ -250,12 +249,12 @@ impl Show for Packed<OutlineElem> {
 
             // Add the overridable outline entry, followed by a line break.
             seq.push(entry.pack());
-            seq.push(LinebreakElem::new().pack());
+            seq.push(LinebreakElem::shared().clone());
 
             ancestors.push(elem);
         }
 
-        seq.push(ParbreakElem::new().pack());
+        seq.push(ParbreakElem::shared().clone());
 
         Ok(Content::sequence(seq))
     }
@@ -272,42 +271,7 @@ impl ShowSet for Packed<OutlineElem> {
 }
 
 impl LocalName for Packed<OutlineElem> {
-    fn local_name(lang: Lang, region: Option<Region>) -> &'static str {
-        match lang {
-            Lang::ALBANIAN => "Përmbajtja",
-            Lang::ARABIC => "المحتويات",
-            Lang::BOKMÅL => "Innhold",
-            Lang::CATALAN => "Índex",
-            Lang::CHINESE if option_eq(region, "TW") => "目錄",
-            Lang::CHINESE => "目录",
-            Lang::CZECH => "Obsah",
-            Lang::DANISH => "Indhold",
-            Lang::DUTCH => "Inhoudsopgave",
-            Lang::ESTONIAN => "Sisukord",
-            Lang::FILIPINO => "Talaan ng mga Nilalaman",
-            Lang::FINNISH => "Sisällys",
-            Lang::FRENCH => "Table des matières",
-            Lang::GERMAN => "Inhaltsverzeichnis",
-            Lang::GREEK => "Περιεχόμενα",
-            Lang::HUNGARIAN => "Tartalomjegyzék",
-            Lang::ITALIAN => "Indice",
-            Lang::NYNORSK => "Innhald",
-            Lang::POLISH => "Spis treści",
-            Lang::PORTUGUESE if option_eq(region, "PT") => "Índice",
-            Lang::PORTUGUESE => "Sumário",
-            Lang::ROMANIAN => "Cuprins",
-            Lang::RUSSIAN => "Содержание",
-            Lang::SERBIAN => "Садржај",
-            Lang::SLOVENIAN => "Kazalo",
-            Lang::SPANISH => "Índice",
-            Lang::SWEDISH => "Innehåll",
-            Lang::TURKISH => "İçindekiler",
-            Lang::UKRAINIAN => "Зміст",
-            Lang::VIETNAMESE => "Mục lục",
-            Lang::JAPANESE => "目次",
-            Lang::ENGLISH | _ => "Contents",
-        }
-    }
+    const KEY: &'static str = "outline";
 }
 
 /// Marks an element as being able to be outlined. This is used to implement the
@@ -363,13 +327,13 @@ impl OutlineIndent {
                             numbering,
                         )?;
 
-                        hidden += numbers + SpaceElem::new().pack();
+                        hidden += numbers + SpaceElem::shared().clone();
                     };
                 }
 
                 if !ancestors.is_empty() {
                     seq.push(HideElem::new(hidden).pack());
-                    seq.push(SpaceElem::new().pack());
+                    seq.push(SpaceElem::shared().clone());
                 }
             }
 
@@ -521,7 +485,7 @@ impl OutlineEntry {
 
 impl Show for Packed<OutlineEntry> {
     #[typst_macros::time(name = "outline.entry", span = self.span())]
-    fn show(&self, _: &mut Engine, _: StyleChain) -> SourceResult<Content> {
+    fn show(&self, _: &mut Engine, styles: StyleChain) -> SourceResult<Content> {
         let mut seq = vec![];
         let elem = self.element();
 
@@ -537,12 +501,31 @@ impl Show for Packed<OutlineEntry> {
             }
         };
 
-        // The body text remains overridable.
+        // Isolate the entry body in RTL because the page number is typically
+        // LTR. I'm not sure whether LTR should conceptually also be isolated,
+        // but in any case we don't do it for now because the text shaping
+        // pipeline does tend to choke a bit on default ignorables (in
+        // particular the CJK-Latin spacing).
+        //
+        // See also:
+        // - https://github.com/typst/typst/issues/4476
+        // - https://github.com/typst/typst/issues/5176
+        let rtl = TextElem::dir_in(styles) == Dir::RTL;
+        if rtl {
+            // "Right-to-Left Embedding"
+            seq.push(TextElem::packed("\u{202B}"));
+        }
+
         seq.push(self.body().clone().linked(Destination::Location(location)));
 
+        if rtl {
+            // "Pop Directional Formatting"
+            seq.push(TextElem::packed("\u{202C}"));
+        }
+
         // Add filler symbols between the section name and page number.
         if let Some(filler) = self.fill() {
-            seq.push(SpaceElem::new().pack());
+            seq.push(SpaceElem::shared().clone());
             seq.push(
                 BoxElem::new()
                     .with_body(Some(filler.clone()))
@@ -550,7 +533,7 @@ impl Show for Packed<OutlineEntry> {
                     .pack()
                     .spanned(self.span()),
             );
-            seq.push(SpaceElem::new().pack());
+            seq.push(SpaceElem::shared().clone());
         } else {
             seq.push(HElem::new(Fr::one().into()).pack());
         }
