@@ -92,6 +92,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
     bed: ['🛏', person: '🛌'],
     bee: '🐝',
     beer: ['🍺', clink: '🍻'],
+    beet: '🫜',
     beetle: ['🪲', lady: '🐞'],
     bell: ['🔔', ding: '🛎', not: '🔕'],
     bento: '🍱',
@@ -435,6 +436,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
         devil.frown: '👿',
         diagonal: '🫤',
         disguise: '🥸',
+        distress: '😫',
         dizzy: '😵',
         dotted: '🫥',
         down: '😞',
@@ -504,7 +506,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
         tear.withheld: '🥹',
         teeth: '😬',
         think: '🤔',
-        tired: '😫',
+        tired: '🫩',
         tongue: '😛',
         tongue.squint: '😝',
         tongue.wink: '😜',
@@ -553,6 +555,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
         front: '🫵',
         m: '🖕',
     ],
+    fingerprint: '🫆',
     fingers: [cross: '🤞', pinch: '🤌', snap: '🫰'],
     fire: '🔥',
     firecracker: '🧨',
@@ -668,6 +671,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
         raised: '🙌',
         shake: '🤝',
     ],
+    harp: '🪉',
     hash: '#',
     hat: [ribbon: '👒', top: '🎩'],
     headphone: '🎧',
@@ -1119,6 +1123,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
     shopping: '🛍',
     shorts: '🩳',
     shoshinsha: '🔰',
+    shovel: '🪏',
     shower: '🚿',
     shrimp: ['🦐', fried: '🍤'],
     shrine: '⛩',
@@ -1163,6 +1168,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
     spider: '🕷',
     spiderweb: '🕸',
     spinach: '🥬',
+    splatter: '🫟',
     sponge: '🧽',
     spoon: '🥄',
     square: [
@@ -1269,6 +1275,7 @@ const EMOJI: &[(&str, Symbol)] = symbols! {
     tree: [
         deciduous: '🌳',
         evergreen: '🌲',
+        leafless: '🪾',
         palm: '🌴',
         xmas: '🎄',
     ],
