@@ -38,32 +38,46 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     ],
 
     // Delimiters.
-    paren: [l: '(', r: ')', t: '⏜', b: '⏝'],
-    brace: [l: '{', r: '}', t: '⏞', b: '⏟'],
+    paren: [l: '(', l.double: '⦅', r: ')', r.double: '⦆', t: '⏜', b: '⏝'],
+    brace: [l: '{', l.double: '⦃', r: '}', r.double: '⦄', t: '⏞', b: '⏟'],
     bracket: [l: '[', l.double: '⟦', r: ']', r.double: '⟧', t: '⎴', b: '⎵'],
-    turtle: [l: '〔', r: '〕', t: '⏠', b: '⏡'],
+    shell: [l: '❲', l.double: '⟬', r: '❳', r.double: '⟭', t: '⏠', b: '⏡'],
     bar: [v: '|', v.double: '‖', v.triple: '⦀', v.broken: '¦', v.circle: '⦶', h: '―'],
     fence: [l: '⧘', l.double: '⧚', r: '⧙', r.double: '⧛', dotted: '⦙'],
     angle: [
         '∠',
         l: '⟨',
-        r: '⟩',
+        l.curly: '⧼',
+        l.dot: '⦑',
         l.double: '《',
+        r: '⟩',
+        r.curly: '⧽',
+        r.dot: '⦒',
         r.double: '》',
         acute: '⦟',
         arc: '∡',
         arc.rev: '⦛',
+        oblique: '⦦',
         rev: '⦣',
         right: '∟',
         right.rev: '⯾',
         right.arc: '⊾',
         right.dot: '⦝',
         right.sq: '⦜',
+        s: '⦞',
         spatial: '⟀',
         spheric: '∢',
         spheric.rev: '⦠',
         spheric.top: '⦡',
     ],
+    ceil: [
+        #[call(crate::math::ceil)] l: '⌈',
+        r: '⌉',
+    ],
+    floor: [
+        #[call(crate::math::floor)] l: '⌊',
+        r: '⌋',
+    ],
 
     // Punctuation.
     amp: ['&', inv: '⅋'],
@@ -84,8 +98,10 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     comma: ',',
     dagger: ['†', double: '‡'],
     dash: [
-        en: '–',
+        #[call(crate::math::accent::dash)] en: '–',
         em: '—',
+        em.two: '⸺',
+        em.three: '⸻',
         fig: '‒',
         wave: '〜',
         colon: '∹',
@@ -93,15 +109,15 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         wave.double: '〰',
     ],
     dot: [
-        op: '⋅',
+        #[call(crate::math::accent::dot)] op: '⋅',
         basic: '.',
         c: '·',
         circle: '⊙',
         circle.big: '⨀',
         square: '⊡',
-        double: '¨',
-        triple: '\u{20db}',
-        quad: '\u{20dc}',
+        #[call(crate::math::accent::dot_double)] double: '¨',
+        #[call(crate::math::accent::dot_triple)] triple: '\u{20db}',
+        #[call(crate::math::accent::dot_quad)] quad: '\u{20dc}',
     ],
     excl: ['!', double: '‼', inv: '¡', quest: '⁉'],
     quest: ['?', double: '⁇', excl: '⁈', inv: '¿'],
@@ -109,15 +125,14 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     hash: '#',
     hyph: ['‐', minus: '\u{2D}', nobreak: '\u{2011}', point: '‧', soft: '\u{ad}'],
     percent: '%',
-    copyright: ['©', sound: '℗'],
     permille: '‰',
     pilcrow: ['¶', rev: '⁋'],
     section: '§',
     semi: [';', rev: '⁏'],
-    slash: ['/', double: '⫽', triple: '⫻', big: '\u{29f8}'],
+    slash: ['/', double: '⫽', triple: '⫻', big: '⧸'],
     dots: [h.c: '⋯', h: '…', v: '⋮', down: '⋱', up: '⋰'],
     tilde: [
-        op: '∼',
+        #[call(crate::math::accent::tilde)] op: '∼',
         basic: '~',
         dot: '⩪',
         eq: '≃',
@@ -133,14 +148,17 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     ],
 
     // Accents, quotes, and primes.
-    acute: ['´', double: '˝'],
-    breve: '˘',
+    acute: [
+        #[call(crate::math::accent::acute)] '´',
+        #[call(crate::math::accent::acute_double)] double: '˝',
+    ],
+    breve: #[call(crate::math::accent::breve)] '˘',
     caret: '‸',
-    caron: 'ˇ',
-    hat: '^',
-    diaer: '¨',
-    grave: '`',
-    macron: '¯',
+    caron: #[call(crate::math::accent::caron)] 'ˇ',
+    hat: #[call(crate::math::accent::hat)] '^',
+    diaer: #[call(crate::math::accent::dot_double)] '¨',
+    grave: #[call(crate::math::accent::grave)] '`',
+    macron: #[call(crate::math::accent::macron)] '¯',
     quote: [
         double: '"',
         single: '\'',
@@ -175,10 +193,12 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         circle.arrow: '⟴',
         circle.big: '⨁',
         dot: '∔',
+        double: '⧺',
         minus: '±',
         small: '﹢',
         square: '⊞',
         triangle: '⨹',
+        triple: '⧻',
     ],
     minus: [
         '−',
@@ -229,17 +249,6 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     gt: [
         '>',
         circle: '⧁',
-        curly: '≻',
-        curly.approx: '⪸',
-        curly.double: '⪼',
-        curly.eq: '≽',
-        curly.eq.not: '⋡',
-        curly.equiv: '⪴',
-        curly.napprox: '⪺',
-        curly.nequiv: '⪶',
-        curly.not: '⊁',
-        curly.ntilde: '⋩',
-        curly.tilde: '≿',
         dot: '⋗',
         approx: '⪆',
         double: '≫',
@@ -250,6 +259,7 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         equiv: '≧',
         lt: '≷',
         lt.not: '≹',
+        neq: '⪈',
         napprox: '⪊',
         nequiv: '≩',
         not: '≯',
@@ -267,17 +277,6 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     lt: [
         '<',
         circle: '⧀',
-        curly: '≺',
-        curly.approx: '⪷',
-        curly.double: '⪻',
-        curly.eq: '≼',
-        curly.eq.not: '⋠',
-        curly.equiv: '⪳',
-        curly.napprox: '⪹',
-        curly.nequiv: '⪵',
-        curly.not: '⊀',
-        curly.ntilde: '⋨',
-        curly.tilde: '≾',
         dot: '⋖',
         approx: '⪅',
         double: '≪',
@@ -288,6 +287,7 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         equiv: '≦',
         gt: '≶',
         gt.not: '≸',
+        neq: '⪇',
         napprox: '⪉',
         nequiv: '≨',
         not: '≮',
@@ -306,11 +306,13 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     prec: [
         '≺',
         approx: '⪷',
+        curly.eq: '≼',
+        curly.eq.not: '⋠',
         double: '⪻',
-        eq: '≼',
-        eq.not: '⋠',
+        eq: '⪯',
         equiv: '⪳',
         napprox: '⪹',
+        neq: '⪱',
         nequiv: '⪵',
         not: '⊀',
         ntilde: '⋨',
@@ -319,11 +321,13 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     succ: [
         '≻',
         approx: '⪸',
+        curly.eq: '≽',
+        curly.eq.not: '⋡',
         double: '⪼',
-        eq: '≽',
-        eq.not: '⋡',
+        eq: '⪰',
         equiv: '⪴',
         napprox: '⪺',
+        neq: '⪲',
         nequiv: '⪶',
         not: '⊁',
         ntilde: '⋩',
@@ -331,10 +335,26 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     ],
     equiv: ['≡', not: '≢'],
     prop: '∝',
+    original: '⊶',
+    image: '⊷',
 
     // Set theory.
-    emptyset: ['∅', rev: '⦰'],
-    nothing: ['∅', rev: '⦰'],
+    emptyset: [
+        '∅',
+        arrow.r: '⦳',
+        arrow.l: '⦴',
+        bar: '⦱',
+        circle: '⦲',
+        rev: '⦰',
+    ],
+    nothing: [
+        '∅',
+        arrow.r: '⦳',
+        arrow.l: '⦴',
+        bar: '⦱',
+        circle: '⦲',
+        rev: '⦰',
+    ],
     without: '∖',
     complement: '∁',
     in: [
@@ -398,7 +418,12 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     ],
 
     // Calculus.
-    infinity: '∞',
+    infinity: [
+        '∞',
+        bar: '⧞',
+        incomplete: '⧜',
+        tie: '⧝',
+    ],
     oo: '∞',
     diff: '∂', // Deprecation planned
     partial: '∂',
@@ -447,7 +472,11 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     // Function and category theory.
     compose: '∘',
     convolve: '∗',
-    multimap: '⊸',
+    multimap: ['⊸', double: '⧟'],
+
+    // Game theory.
+    tiny: '⧾',
+    miny: '⧿',
 
     // Number theory.
     divides: ['∣', not: '∤'],
@@ -456,7 +485,18 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     wreath: '≀',
 
     // Geometry.
-    parallel: ['∥', circle: '⦷', not: '∦'],
+    parallel: [
+        '∥',
+        struck: '⫲',
+        circle: '⦷',
+        eq: '⋕',
+        equiv: '⩨',
+        not: '∦',
+        slanted.eq: '⧣',
+        slanted.eq.tilde: '⧤',
+        slanted.equiv: '⧥',
+        tilde: '⫳',
+    ],
     perp: ['⟂', circle: '⦹'],
 
     // Miscellaneous Technical.
@@ -479,19 +519,75 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
     yen: '¥',
 
     // Miscellaneous.
-    ballot: ['☐', x: '☒'],
-    checkmark: ['✓', light: '🗸'],
+    ballot: ['☐', cross: '☒', check: '☑', check.heavy: '🗹'],
+    checkmark: ['✓', light: '🗸', heavy: '✔'],
+    crossmark: ['✗', heavy: '✘'],
     floral: ['❦', l: '☙', r: '❧'],
-    notes: [up: '🎜', down: '🎝'],
     refmark: '※',
-    servicemark: '℠',
+    copyright: ['©', sound: '℗'],
+    copyleft: '🄯',
+    trademark: ['™', registered: '®', service: '℠'],
     maltese: '✠',
-    suit: [club: '♣', diamond: '♦', heart: '♥', spade: '♠'],
+    suit: [
+        club.filled: '♣',
+        club.stroked: '♧',
+        diamond.filled: '♦',
+        diamond.stroked: '♢',
+        heart.filled: '♥',
+        heart.stroked: '♡',
+        spade.filled: '♠',
+        spade.stroked: '♤',
+    ],
+
+    // Music.
+    note: [
+        up: '🎜',
+        down: '🎝',
+        whole: '𝅝',
+        half: '𝅗𝅥',
+        quarter: '𝅘𝅥',
+        quarter.alt: '♩',
+        eighth: '𝅘𝅥𝅮',
+        eighth.alt: '♪',
+        eighth.beamed: '♫',
+        sixteenth: '𝅘𝅥𝅯',
+        sixteenth.beamed: '♬',
+        grace: '𝆕',
+        grace.slash: '𝆔',
+    ],
+    rest: [
+        whole: '𝄻',
+        multiple: '𝄺',
+        multiple.measure: '𝄩',
+        half: '𝄼',
+        quarter: '𝄽',
+        eighth: '𝄾',
+        sixteenth: '𝄿',
+    ],
+    natural: [
+        '♮',
+        t: '𝄮',
+        b: '𝄯',
+    ],
+    flat: [
+        '♭',
+        t: '𝄬',
+        b: '𝄭',
+        double: '𝄫',
+        quarter: '𝄳',
+    ],
+    sharp: [
+        '♯',
+        t: '𝄰',
+        b: '𝄱',
+        double: '𝄪',
+        quarter: '𝄲',
+    ],
 
     // Shapes.
     bullet: '•',
     circle: [
-        stroked: '○',
+        #[call(crate::math::accent::circle)] stroked: '○',
         stroked.tiny: '∘',
         stroked.small: '⚬',
         stroked.big: '◯',
@@ -509,33 +605,33 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         filled.v: '⬮',
     ],
     triangle: [
-        stroked.r: '▷',
-        stroked.l: '◁',
         stroked.t: '△',
         stroked.b: '▽',
+        stroked.r: '▷',
+        stroked.l: '◁',
         stroked.bl: '◺',
         stroked.br: '◿',
         stroked.tl: '◸',
         stroked.tr: '◹',
-        stroked.small.r: '▹',
-        stroked.small.b: '▿',
-        stroked.small.l: '◃',
         stroked.small.t: '▵',
+        stroked.small.b: '▿',
+        stroked.small.r: '▹',
+        stroked.small.l: '◃',
         stroked.rounded: '🛆',
         stroked.nested: '⟁',
         stroked.dot: '◬',
-        filled.r: '▶',
-        filled.l: '◀',
         filled.t: '▲',
         filled.b: '▼',
+        filled.r: '▶',
+        filled.l: '◀',
         filled.bl: '◣',
         filled.br: '◢',
         filled.tl: '◤',
         filled.tr: '◥',
-        filled.small.r: '▸',
-        filled.small.b: '▾',
-        filled.small.l: '◂',
         filled.small.t: '▴',
+        filled.small.b: '▾',
+        filled.small.r: '▸',
+        filled.small.l: '◂',
     ],
     square: [
         stroked: '□',
@@ -576,14 +672,19 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         filled.small: '⬪',
         filled.medium: '⬧',
     ],
+    parallelogram: [
+        stroked: '▱',
+        filled: '▰',
+    ],
     star: [op: '⋆', stroked: '☆', filled: '★'],
 
     // Arrows, harpoons, and tacks.
     arrow: [
-        r: '→',
+        #[call(crate::math::accent::arrow)] r: '→',
         r.long.bar: '⟼',
         r.bar: '↦',
         r.curve: '⤷',
+        r.turn: '⮎',
         r.dashed: '⇢',
         r.dotted: '⤑',
         r.double: '⇒',
@@ -607,9 +708,10 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         r.twohead.bar: '⤅',
         r.twohead: '↠',
         r.wave: '↝',
-        l: '←',
+        #[call(crate::math::accent::arrow_l)] l: '←',
         l.bar: '↤',
         l.curve: '⤶',
+        l.turn: '⮌',
         l.dashed: '⇠',
         l.dotted: '⬸',
         l.double: '⇐',
@@ -637,6 +739,7 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         t: '↑',
         t.bar: '↥',
         t.curve: '⤴',
+        t.turn: '⮍',
         t.dashed: '⇡',
         t.double: '⇑',
         t.filled: '⬆',
@@ -648,6 +751,7 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         b: '↓',
         b.bar: '↧',
         b.curve: '⤵',
+        b.turn: '⮏',
         b.dashed: '⇣',
         b.double: '⇓',
         b.filled: '⬇',
@@ -656,7 +760,7 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         b.stroked: '⇩',
         b.triple: '⤋',
         b.twohead: '↡',
-        l.r: '↔',
+        #[call(crate::math::accent::arrow_l_r)] l.r: '↔',
         l.r.double: '⇔',
         l.r.double.long: '⟺',
         l.r.double.not: '⇎',
@@ -715,13 +819,13 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         b: '⌄',
     ],
     harpoon: [
-        rt: '⇀',
+        #[call(crate::math::accent::harpoon)] rt: '⇀',
         rt.bar: '⥛',
         rt.stop: '⥓',
         rb: '⇁',
         rb.bar: '⥟',
         rb.stop: '⥗',
-        lt: '↼',
+        #[call(crate::math::accent::harpoon_lt)] lt: '↼',
         lt.bar: '⥚',
         lt.stop: '⥒',
         lb: '↽',
@@ -784,106 +888,109 @@ pub(crate) const SYM: &[(&str, Symbol)] = symbols! {
         l.r: '⟛',
     ],
 
-     // Lowercase Greek.
-     alpha: 'α',
-     beta: ['β', alt: 'ϐ'],
-     chi: 'χ',
-     delta: 'δ',
-     epsilon: ['ε', alt: 'ϵ'],
-     eta: 'η',
-     gamma: 'γ',
-     iota: 'ι',
-     kai: 'ϗ',
-     kappa: ['κ', alt: 'ϰ'],
-     lambda: 'λ',
-     mu: 'μ',
-     nu: 'ν',
-     ohm: ['Ω', inv: '℧'],
-     omega: 'ω',
-     omicron: 'ο',
-     phi: ['φ', alt: 'ϕ'],
-     pi: ['π', alt: 'ϖ'],
-     psi: 'ψ',
-     rho: ['ρ', alt: 'ϱ'],
-     sigma: ['σ', alt: 'ς'],
-     tau: 'τ',
-     theta: ['θ', alt: 'ϑ'],
-     upsilon: 'υ',
-     xi: 'ξ',
-     zeta: 'ζ',
+    // Lowercase Greek.
+    alpha: 'α',
+    beta: ['β', alt: 'ϐ'],
+    chi: 'χ',
+    delta: 'δ',
+    epsilon: ['ε', alt: 'ϵ'],
+    eta: 'η',
+    gamma: 'γ',
+    iota: 'ι',
+    kai: 'ϗ',
+    kappa: ['κ', alt: 'ϰ'],
+    lambda: 'λ',
+    mu: 'μ',
+    nu: 'ν',
+    ohm: ['Ω', inv: '℧'],
+    omega: 'ω',
+    omicron: 'ο',
+    phi: ['φ', alt: 'ϕ'],
+    pi: ['π', alt: 'ϖ'],
+    psi: 'ψ',
+    rho: ['ρ', alt: 'ϱ'],
+    sigma: ['σ', alt: 'ς'],
+    tau: 'τ',
+    theta: ['θ', alt: 'ϑ'],
+    upsilon: 'υ',
+    xi: 'ξ',
+    zeta: 'ζ',
 
-     // Uppercase Greek.
-     Alpha: 'Α',
-     Beta: 'Β',
-     Chi: 'Χ',
-     Delta: 'Δ',
-     Epsilon: 'Ε',
-     Eta: 'Η',
-     Gamma: 'Γ',
-     Iota: 'Ι',
-     Kai: 'Ϗ',
-     Kappa: 'Κ',
-     Lambda: 'Λ',
-     Mu: 'Μ',
-     Nu: 'Ν',
-     Omega: 'Ω',
-     Omicron: 'Ο',
-     Phi: 'Φ',
-     Pi: 'Π',
-     Psi: 'Ψ',
-     Rho: 'Ρ',
-     Sigma: 'Σ',
-     Tau: 'Τ',
-     Theta: 'Θ',
-     Upsilon: 'Υ',
-     Xi: 'Ξ',
-     Zeta: 'Ζ',
+    // Uppercase Greek.
+    Alpha: 'Α',
+    Beta: 'Β',
+    Chi: 'Χ',
+    Delta: 'Δ',
+    Epsilon: 'Ε',
+    Eta: 'Η',
+    Gamma: 'Γ',
+    Iota: 'Ι',
+    Kai: 'Ϗ',
+    Kappa: 'Κ',
+    Lambda: 'Λ',
+    Mu: 'Μ',
+    Nu: 'Ν',
+    Omega: 'Ω',
+    Omicron: 'Ο',
+    Phi: 'Φ',
+    Pi: 'Π',
+    Psi: 'Ψ',
+    Rho: 'Ρ',
+    Sigma: 'Σ',
+    Tau: 'Τ',
+    Theta: 'Θ',
+    Upsilon: 'Υ',
+    Xi: 'Ξ',
+    Zeta: 'Ζ',
 
-     // Hebrew.
-     aleph: 'א',
-     alef: 'א',
-     beth: 'ב',
-     bet: 'ב',
-     gimmel: 'ג',
-     gimel: 'ג',
-     daleth: 'ד',
-     dalet: 'ד',
-     shin: 'ש',
+    // Hebrew.
+    // In math, the following symbols are replaced with corresponding characters
+    // from Letterlike Symbols.
+    // See https://github.com/typst/typst/pull/3375.
+    aleph: 'א',
+    alef: 'א',
+    beth: 'ב',
+    bet: 'ב',
+    gimmel: 'ג',
+    gimel: 'ג',
+    daleth: 'ד',
+    dalet: 'ד',
+    shin: 'ש',
 
-     // Double-struck.
-     AA: '𝔸',
-     BB: '𝔹',
-     CC: 'ℂ',
-     DD: '𝔻',
-     EE: '𝔼',
-     FF: '𝔽',
-     GG: '𝔾',
-     HH: 'ℍ',
-     II: '𝕀',
-     JJ: '𝕁',
-     KK: '𝕂',
-     LL: '𝕃',
-     MM: '𝕄',
-     NN: 'ℕ',
-     OO: '𝕆',
-     PP: 'ℙ',
-     QQ: 'ℚ',
-     RR: 'ℝ',
-     SS: '𝕊',
-     TT: '𝕋',
-     UU: '𝕌',
-     VV: '𝕍',
-     WW: '𝕎',
-     XX: '𝕏',
-     YY: '𝕐',
-     ZZ: 'ℤ',
+    // Double-struck.
+    AA: '𝔸',
+    BB: '𝔹',
+    CC: 'ℂ',
+    DD: '𝔻',
+    EE: '𝔼',
+    FF: '𝔽',
+    GG: '𝔾',
+    HH: 'ℍ',
+    II: '𝕀',
+    JJ: '𝕁',
+    KK: '𝕂',
+    LL: '𝕃',
+    MM: '𝕄',
+    NN: 'ℕ',
+    OO: '𝕆',
+    PP: 'ℙ',
+    QQ: 'ℚ',
+    RR: 'ℝ',
+    SS: '𝕊',
+    TT: '𝕋',
+    UU: '𝕌',
+    VV: '𝕍',
+    WW: '𝕎',
+    XX: '𝕏',
+    YY: '𝕐',
+    ZZ: 'ℤ',
 
-     // Miscellaneous letter-likes.
-     ell: 'ℓ',
-     planck: ['ℎ', reduce: 'ℏ'],
-     angstrom: 'Å',
-     kelvin: 'K',
-     Re: 'ℜ',
-     Im: 'ℑ',
-     dotless: [i: '𝚤', j: '𝚥'],
+    // Miscellaneous letter-likes.
+    ell: 'ℓ',
+    planck: ['ℎ', reduce: 'ℏ'],
+    angstrom: 'Å',
+    kelvin: 'K',
+    Re: 'ℜ',
+    Im: 'ℑ',
+    dotless: [i: '𝚤', j: '𝚥'],
 };
